%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% convex-Asm.SL - convex specific inFormation for LAP-TO-Asm            %
%                                                                       %
% Author:                                                               %
% Date:                                                                 %
%                                                                       %
% Copyright (c) 1988 Konrad Zuse Zentrum Berlin                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(CodeFileNameFormat*
         DataFileNameFormat*
         DoubleFloatFormat*
%        GlobalDataFileName*
         CommentFormat*
         LabelFormat*
         SpecialLabelFormat*
         LocalLabelS*
         ExternalDeclarationFormat*
         ExportedDeclarationFormat*
         FullwordFormat*
         HalfwordFormat*
         ReserveDataBlockFormat*
         ReserveZeroBlockFormat*
         DefinedFunctionCellFormat*
         UndefinedFunctionCellinstructionS*
         MainEntryPointName*
         CodeOut*
         DataOut*
         *counting
         *Lower
         *DeclareBeforeUse
         AsmOpenparen*
         AsmCloseparen*
         NumericRegisterNameS*
         ForeignExternList*
         asmcomment*
         nilnumber*
         DataProcState*))
 
(compiletime (load sys-consts))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This assembler produces two Files: a Code File and a Data File. Both  %
% of these Files are assembled to reside in the convex Dataspace        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  Formats                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(setq CodeFileNameFormat* "%w.s")
           % CAL doesnt like .xxx
(setq DataFileNameFormat* "%wd.s")
 
 
% Use %e in DoubleFloatFormat* to be "safe", in case the argument passed
% is something that evaluates to a Float, rather than a Float itself.  ALSO
% note that this will only work if the cross compiler is running on a
% machine that Prints Floats in a Format that the convex will accept. This
% appears to be true on the Unix cross compiler, but may not be true on the
% DEC-20.  (Consider the case where the Float is 0.1 )
 
(setq DoubleFloatFormat* "     ds.f %e%n")
 
%-----------------------------------------------------------------------
%-- rrk 11/11/82 - The spaces before the instruction is Printed.
(setq instructionLeadingSpaces* "     ")
%-----------------------------------------------------------------------
 
(setq SpecialLabelFormat* "%w: ")  % Labels are in the first column
(setq LabelFormat* "%w: ")         % Labels are in the first column
 
% Define cal parens
(setq AsmOpenparen* "")
(setq AsmCloseparen* "")
 
%  Redefine DataPrintLabel, so we can fix the problem with Data Labels
%   occuring on the wrong line.
(de DataPrintLabel (LAB)
  (when (member lab '(SYMVAL SYMNAM SYMPRP SYMFNC SYMGET))
        (DataPrintF "        .align 8 %n" nil nil))
  (DataPrintF labelformat* LAB lab))
 
 
(setq CommentFormat* "; %p%n")        % Comments begin with a semicolon
 
(setq ExportedDeclarationFormat* "     .globl %w%n")
 
(setq ExternalDeclarationFormat* "     .globl %w%n")
 
(setq *DeclarebeforeUse T)            % Force Early Entry?
 
(setq FullwordFormat* "     ds.w   %e%n") 
%/HalfwordFormat* :="     CON %e%n"$   % Will EVAL Formatter
 
(setq ReserveDataBlockFormat* "     bs.w  %e%n")  %/ Bytes, so divide by 8?
(setq ReserveZeroBlockFormat* "%w:%n       bs.w %e%n")
 
 
(put 'Mkitem 'AsmExpressionFormat "%e*TAGSHIFT+%e") % MkItem(Tag, inf)
%/ need to define Tagshift
 
(setq DefinedFunctionCellFormat* "     ds.w %w%n")
 
(setq UndefinedFunctionCellinstructions*
    % '((callq (plus2 symfnc (times2 6 (idloc UndefinedFunction)))))
    %  (list "ds.w    " (get 'UndefinedFunction 'symbol))
       '((ds!.w   UndefinedFunction))
   )
 
(de dataprintundefinedfunctioncell ()
 (let ((x (get 'undefinedfunction 'entrypoint)))
  (dataprintf definedfunctioncellformat* x x)))
 
(setq LAMBDAFunctionCellinstructionS*
      '((callq COMPILEDCALLINGINTERPRETED)) )
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%            ListS and ConstANT DEFinITIONS                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(put 'symfnc 'symbol 'symfnc) (put 'symval 'symbol 'symval)
(put 'symprp 'symbol 'symprp) (put 'symnam 'symbol 'symnam)
 
(setq MainEntryPointName* '_main)  % chose a simple default
                                       % Main procedure Name
 
(setq NumericRegisterNameS* '[ nil "s0" "s1" "s2" "s3" "s4" "s5" "s6" "s7"])
  % Keep some Si free
 
%/AsmOpenparen* := '($    %/ No parens
%/AsmCloseparen* := ')$
 
%DefList('((Land &)     %/ none legal in CAL
%        (Lor  )
%        (LSH <<)), 'BinaryAsmOp)$
 
 
(deFList '((A0 "sp") (A1 "a1") (A2 "a2") (A3 "a3") (A4 "a4") (A5 "a5")
           (A6 "ap") (A7 "fp"))
         'RegisterName)
 
(DefList '((nil "s5")                     % This reg always holds NIL.     
  (t0 "a1") (t1 "a2") (t2 "a3") (t3 "s6") (t4 "s7") (t5 "a4") (t6 "a5")
  (sp "sp") (ap "ap") )
          'RegisterName)

 
(deFList '((S0 "s0") (S1 "s1") (S2 "s2") (S3 "s3") (S4 "s4") (S5 "s5")
           (S6 "s6") (S7 "s7"))
         'RegisterName)
 
(deFList '((V0 "v0") (V1 "v1") (V2 "v2") (V3 "v3") (V4 "v4") (V5 "v5")
           (V6 "v6") (V7 "v7") (VL "vl") (VM "vm") (VLS "vls") (VS "vs")
          )
         'RegisterName)
 
(setq DataProcState* 'Data)
 
% This patch is until init Files can be read
 
%(COPYD 'SAVEUNCOMPILEDExpression 'SAVEForCOMPILATION)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%            Code HEADERS and TRAILERS                                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(de CodeFileHEADER NIL        % Pure Code Segment
     (prog (init-size)                   % Must be loaded to 200(8)
         (CodePrintF "     .fpmode native%n")
         (setq DataProcState* 'Data)
         (CodePrintF "; this is the start of execution%n")
  %/ Issue EXT early
         (CodeDeclareExternal 'SYMVAL)
         (CodeDeclareExternal 'SYMfnC)
         (CodeDeclareExternal 'SYMPRP)
         (CodeDeclareExternal 'SYMNAM)
         (CodePrintF "TAGSHIFT = %w%n" (expt 2 26))
              % prepare memory allocation
       (setq init-size (plus (compiler-Constant 'initial-stacksize)
                             (compiler-Constant 'initial-heapsize)
                             (compiler-Constant 'initial-bndstksize)
                             (compiler-Constant 'initial-bpssize) 30))
        (CodePrintf "     .text         %n")
%       (CodePrintf "     .globl _main  %n")
%       (CodePrintf "_main:             %n")
%       (CodePrintf "     .globl inTALLC      ; does the inital alloc %n")
%       (CodePrintf "     sub.w #16,sp  %n")
%       (CodePrintf "     pshea #%d     %n" init-size)
%       (CodePrintf "     mov.w  sp,ap  %n")
%       (CodePrintf "     pshea #1      %n")
%       (CodePrintf "     calls inTALLC %n")
%       (CodePrintf "     add.w #8,sp   %n")
              % load nil register
%       (CodePrintF " ;   st.l s0,carNIL %n")                          
%       (CodePrintF " ;   st.l s0,carNIL+1 %n")% car and cdr of nil   
              % initial start
%       (CodePrintf "     callq _psl_main %n")
%       (CodePrintF "            ;now return to OS%n")
%       (CodePrintf "     mov.w  sp,ap  %n")
%       (CodePrintf "     mov.w  sp,ap  %n")
%       (CodePrintf "     pshea #0      %n")
%       (CodePrintf "     calls _exit   %n")
%       (CodePrintf "     add.w #4,sp   %n")
%       (CodePrintF "    rtn %n")
        (setq ForeignExternList* NIL)
         NIL))
 
(de DataFileHEADER NIL
            (DataPrintf "     .data         %n")
            (DataPrintF "TAGSHIFT = %w%n" (expt 2 26))
            (DataPrintf "     .globl _data_start  %n")
            (DataPrintf "_data_start:    ds.w   _data_start%n")
            (DataPrintf "     .globl _data_end  %n")
            (DataPrintf "_data_end:      ds.w   _data_segment_end%n")
           )

 
(de DataFileTRAILER nil 
            (DataPrintf "_data_segment_end:  ds.w  0%n")
         )
 
(de CodeFileTRAILER NIL nil)
 
(de CHECKForeignEXTERN (fn)
  (cond ((not (memq fn ForeignExternList*))
         (setq ForeignExternList* (CONS fn ForeignExternList*)))))
 
(de CodeBlockHEADER NIL  %/ Chuck this?
  NIL)
 
(de CodeBlockTRAILER NIL
  NIL)
 
(de DataALIGNFullword NIL
  (DataPrintF "     .align  2%n"))
 
(de CodeALIGNFullword NIL 
  (CodePrintF "     .align  2%n"))
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                   Print PROCEDURES                                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de PrintString (S)
  (prog (N N8)
        (prin2 "     ds.b ")
        (prin2 '!")
        (setq N (sub1 (string-length S)))
        (cond ((neq N -1)
                      (For (FROM I 0 N 1)
                           (DO (Prin2b (strbyt(strinf S) I ))))))
        (prin2 "\0") (prin2t '!") 
        (prin2t " .align 2") % align follwing data 
   ))
 
(de truncateString (S N)
  (cond ((leq (size S) N) S)
        (T (sub S 0 N))))
 
 
(de PrintByteList (L)
  (prog (N N8)
        (prin2 "     ds.b ")
        (prin2 '!")
        (foreach X in L DO (prin2b X))
        (prin2t '!") ))
 
(de prin2b (x)
    (cond ((and (numberp x)(or (lessp x 32) (greaterp x 127)))
             (prin2 "\")(prin2oct x t))
          ((eq x (id2int '!") ) (prin2 "\")(prin2oct x t))
          ((numberp x)(prin2 (int2id x)))
          (t (prin2 x)) ))

(de prin2oct(x c)
    (cond ((eq x 0)(when c (prin2 0)))
          (t (prin2oct (quotient x 8) nil)(prin2 (remainder x 8)))))
 
(de PrintByte (X)                   %/ careful, not packed?
         (prin2 "     ds.b ")
         (PrintExpression X)
         (terpri))
 
 
(de PrintHalfwordList (L)
  (foreach X in L DO (PrintHalfword X)))
 
 
(de PrintHalfword (X)
         (prin2 "     ds.w ")
         (PrintExpression X)
         (terpri))
 
 
(de PrintHalfwordS (X)
         (rplaca PrintExpressionFormPointER* X)
         (PrintF HalfwordFormat* PrintExpressionForm*))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% AsmPseudoOp, handles OPCode Specially, rather than as just
% instruction Output Form is:
% "tab" <opCode>   "space" <Operand> { "comma" <Operand> }   "newline"
 
(put 'Comment 'AsmPseudoOP 'PrintComment)
 
(put 'Halfword 'AsmPseudoOP 'AsmPseudoPrintHalfwordS)
 
(de AsmPseudoPrintHalfwordS (X)
  (foreach Y in (cdr X) DO (PrintHalfword Y)))
 
 
(de PrintOPCode (X)       % default OpCode Printer
        (setq *lower T)   % Kleinbuchstaben fuer opCode
        (prin2 X)
        (setq *lower NIL) )
 
(de SpecialActionFormainEntryPoint NIL
         (CodeAlignFullword)
         (cond (*DeclareBEForEUse (CodeDeclareExported
                                    MainEntryPointName*)))
         (CodePrintLabel MainEntryPointName*)
         (cond ((not *DeclareBEForEUse)
                (CodeDeclareExported MainEntryPointName*)))
         NIL) 
 
(de AsmSymbolp (X)
  (convexSymbolp (cond ((IDP X) (ID2String X))
                     (T X))))
 
 
(de convexSymbolp (X)
%/ do not Use @, confuzed with getMEM, ie as NENAM, not OP...
  (prog (N C I)
        (setq N (size X))
        (cond ((geq N 8) (return NIL)))
        (setq C (inDX X 0))
        (cond ((not (or (and (geq C (char A)) (leq C (char Z)))
                        (equal C (char !%)) (equal C (char $))))
               (return NIL)))
        (setq I 1)
   loop
        (cond ((greaterp I N) (return T)))
        (setq C (inDX X I))
        (cond ((not (or (and (geq C (char A)) (leq C (char Z)))
                        (equal C (char !%)) (equal C (char $))
                        (and (geq C (char 0)) (leq C (char 9)))))
               (return NIL)))
        (setq I (plus I 1))
        (go loop)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% OperandPrintFunctions:
% You have to be careful with the adressing modes on the convex regs.
% Only A regs can be indexes, cant touch A0, S0, etc
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de PrintNumericOperand (X)
  (prin2 "#")
  (prin2 X))
 
 
(de OperandPrintindexed (X)
  (prog (RG DISP)
        (setq X (cdr X))
        (setq RG (car X))
        (setq DISP (cadr X))
%/ Check that AREGP and Number/AddressP
        (PrintExpression DISP)
        (prin2 "(")
        (PrintOperand RG)
        (prin2 ")")
        ))
 
(put 'indexed 'OperandPrintFunction 'OperandPrintindexed)
 
(de OperandPrintImmediate (X) % (Immediate x) % for ADDRESS
  (setq x (cadr x))
  (when (not (eqcar x 'entry)) (prin2 "#"))
  (PrintExpression  X))
 
(put 'Immediate 'OperandPrintFunction 'OperandPrintImmediate)
 
(de OperandPrintLabelGEN (X)                  % for ADDRESS
  (PrintExpression (APPLY (car X) (cdr X))))
 
(put 'LabelGEN 'OperandPrintFunction 'OperandPrintLabelGEN)
 
(de OperandPrintIConst (X)
  (PrintExpression (cadr X)))
 
(put 'IConst 'OperandPrintFunction 'OperandPrintIConst)
 
(de OperandPrintForeignEntry (X)  % (ForeignEntry FcnName)
     (prog (*lower)
        (setq *lower t)
        (prin2 "_") (prin2 (cadr X))
        (CHECKForeignEXTERN (cadr X))))
 
(put 'ForeignEntry 'OperandPrintFunction 'OperandPrintForeignEntry)
 
(put 'ForeignEntry 'AsmExpressionFunction 'OperandPrintForeignEntry)
 
(de OperandPrintComment (X)       % (Comment ...)
        (PrintComment (cadr X)))
 
(put 'Comment 'OperandPrintFunction 'OperandPrintComment)
 
(de MCPrint (X)  % Echo of MC's
  (cond ((EQcar X '*JUMPON) (CodePrintF ";       Jumpon cmacro %n"))
        ((eqcar x 'STRING)  (CodePrintF ";       (String .... )%n "))
        (T (CodePrintF ";       %p%n" X))))
 
(de instructionPrint (X)
  (cond ((eqcar x 'STRING)  (CodePrintF ";                 (String .... )%n"))
        (t                  (CodePrintF ";                 %p%n" X))))
 
(put 'CERRor 'AsmPseudoOP 'PrintComment)
 
% Use the flag to tell if we have two Labels in a row.  If so, generate
%  the Special syntax of Label:    for the first Label and continue
%  normally for the second.
 
(Global '(*Labelfound))
 
% Variable with the number of links in the Code body.
 
(fluid '(&LinKS-present&))
 
(de AsmOutLAP (U)
  (prog (LocalLabelS* oldOut)
    %   (setq &LinKS-present& % look-for-links is in CRlap
    %         (or (look-for-links U) %%( assoc '*LAMBinD U)
    %             (assoc '*Id-Apply U)
    %     %%%%    (assoc '*progBinD U)
    %             (assoc '*call U)))
        (setq U (pass1lap U))          % Expand cmacros, quoted Expressions
        (setq U (alignall U))
        (CodeBlockHeader)
        (setq oldout (WRS CodeOut*))
        (setq *Labelfound NIL)        % Reset to NIL in the beginning.
        (foreach X in u do (AsmOutLAP1 X))
        (wrs oldout)
        (CodeBlockTrailer)))
 
(de alignall (u)
    % insert .align pseudos in front of all fullwords and labels to fullwords
   (let (r)
      (while u
         (cond((eqcar (car u) 'FULLWORD) (push '(!.align! 2) r))
              ((and (atom (car u)) (eqcar (cadr u) 'FULLWORD))
               (push '(!.align! 2) r)
               (push (pop u) r)))  
         (push (pop u) r))
      (reversip r) ))
 
%-- rrk 11/11/82 Modified this Function becaUse it is not putting Out enough
%    spaces at the front of the line.  It has a variable,
%    instructionLeadingSpaces* that must be set to the number of spaces to
%    Print at the front of the line for each new instruction.
 
% Modify PrintLabel, so it will check the Labelfound flag and if already on
%  will Print the Special stuff on the end.
 
(de convexPrintLabel (X)
  (cond (*Labelfound (PrintF SpecialLabelFormat* X X))
        (T (progn (setq *Labelfound T)
                  (PrintF LabelFormat* X X)))))
 
 
 
(de CodeDeclareExportedUse (Y)
  (cond (*DeclareBEForEUse (progn (CodeDeclareExported Y)
                                   (convexPrintLabel Y)))
        (T (progn (convexPrintLabel Y)
                  (CodeDeclareExported Y)))))
 

% Support for SYMGET
%
%
 
(de initializesymboltable ()
  (let ((maxsymbol (compiler-constant 'maxsymbols)))
    (when (lessp maxsymbol nextidnumber*)
      (errorprintf "*** MaxSymbols %r is too small; at least %r are needed"
                   maxsymbol nextidnumber*)
      (setq maxsymbol (plus nextidnumber* 100)))
    (flag '(nil) 'nilinitialvalue)
    (put 't 'initialvalue 't)
    (setq nilnumber* (compileconstant nil))
    (dataalignfullword)
    (initializesymval)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymprp)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymnam)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymfnc)
    (datareservefunctioncellblock
     (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymget)   % SYMGET feature
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    ))
 
(flag '(initializesymboltable) 'lose)
 
(de initializesymget ()
  (dataprintgloballabel (findgloballabel 'symget))
  (for (from i 0 128 1) (do
                              (dataprintfullword nilnumber*)))
  (foreach x in (car orderedidlist*) do
                              (dataprintfullword nilnumber*))
  )

(de codeprintfullword (x)
  (rplaca printexpressionformpointer* x)
  (CodeALIGNFullword)
  (codeprintf fullwordformat* printexpressionform*))
 
(flag '(codeprintfullword) 'lose)


