%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File: VECTOR-INSTRUCTIONS
%
% Description:   Vector instruction interface for Convex C-1:
%                Vector instructions, vector Cmacros
%
% Date:          19 Oct 1988
% Author:        W. Neun, ZIB Berlin
%
% Copyright (c) 1988, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(*VectorInstructions*))   % list of all vector instrs
 
(loadtime (load compiler))
 
(compiletime (load big-faslend))
 
% associate with a number 0-7 a vector register and a vectorlength reg
 
(de vregname (x)  (nth '((reg v0) (reg v1) (reg v2) (reg v3)
                         (reg v4) (reg v5) (reg v6) (reg v7))
                         (wplus2 x 1) ))
(put 'vregname 'operandvaluefunction 'vregname)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(put 'vector1pat 'pattern '(nil nil (any (mac a1))))
 
(put 'vector1valuepat 'pattern '(nil nil (any (mac a1 dest))))
 
(put 'vector2pat 'pattern '(nil nil (any (mac a1 a2))))
 
(put 'vector2valuepat 'pattern '(nil nil (any (mac a1 a2 dest))))
 
(put 'vector3pat 'pattern '(nil nil (any (mac a1 a2 a3))))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  definition of vector op codes, cmacros and
%  interface functions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  load, store, element handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%----------------------------------------------------------------------%
% (vstore (vreg 3) (loc (getmem x)) 4)
% store from vectorregister v3 to memory with offset 4
%----------------------------------------------------------------------%
 
(put 'vstore 'openfn '(vector3pat *vstore))
 
(de *vstore (a1 a2 a3)
        (expand3operandcmacro a1 a2 a3 '*vstore ))
 
(defcmacro *vstore
    ((numberp anyp onep) (*move argtwo (reg a6))
                         (ld!.w 1 (reg vs))
                         (st!.l (vregname argone) (reg A6)))
    ((numberp anyp anyp) (*move argtwo (reg a6))
                         (*move argthree (reg a7))
                         (mov (reg a7) (reg vs))
                         (st!.l (vregname argone) (reg A6)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vscatter (vreg 3) (loc (getmem x)) (vreg 4))
% store from vectorregister v3 to memory (reg a5)  vreg 4
%----------------------------------------------------------------------%
 
(put 'vscatter 'openfn '(vector3pat *vscatter))
 
(de *vscatter (a1 a2 a3)
        (expand3operandcmacro a1 a2 '*vscatter )))
 
(defcmacro *vscatter
    ((numberp any numberp) (*move argtwo (reg a5)) % MUST be A5
                       (stvi!.l (vregname argone) (vregname argthree)))
    (                  (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vload (vreg 3) (loc (getmem x)) 4)
% store from vectorregister v3 to memory with offset 4
%----------------------------------------------------------------------%
 
(put 'vload  'openfn '(vector3pat *vload ))
 
(de *vload  (a1 a2 a3)
        (expand3operandcmacro a1 a2 a3 '*vload))
 
(defcmacro *vload
    ((numberp anyp onep) (*move argtwo (reg a6))
                         (ld!.w 1 (reg vs))
                         (ld!.l (reg A6) (vregname argone)))
    ((numberp anyp anyp) (*move argtwo (reg a6))
                         (*move argthree (reg a7))
                         (mov (reg a7) (reg vs))
                         (ld!.l (vregname argone) (reg A6)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vgather (vreg 3) (loc (getmem x)) (vreg 4))
% store from vectorregister v3 to memory with offset vreg 4
%----------------------------------------------------------------------%
 
(put 'vgather 'openfn '(vector3pat *vgather ))
 
(de *vgather (a1 a2 a3)
        (expand3operandcmacro a1 a2 a3 '*vgather )))
 
(defcmacro *vgather
    ((numberp anyp numberp) (*move argtwo (reg a5))
                       (ldvi!.l (vregname argthree) (vregname argone)))
    (                  (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VsetElement (vreg 3) n x)
% set element n of v3 to value x
%----------------------------------------------------------------------%
 
(put 'VsetElement  'openfn '(vector3pat *vsetel))
 
(de *vsetel (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vsetel))
 
(defcmacro *vsetel
    ((numberp Sregp sregp)
                         (mov argthree argtwo (vregname argone)))
    ((numberp anyp sregp)
                         (*move argtwo (reg s6))
                         (mov argthree (reg S6) (vregname argone)))
    ((numberp anyp anyp)
                         (*move argtwo (reg s7))
                         (*move argthree (reg s6))
                         (*vsetel argone (reg s7)(reg s6)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VgetElement (vreg 3) n)
% get element n of v3 (function value)
%----------------------------------------------------------------------%
 
(put 'VgetElement  'openfn '(vector2valuepat *vgetel))
 
(de *vgetel (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vgetel))
 
(defcmacro *vgetel
    ((numberp Sregp sregp)
                         (mov (vregname argone) argtwo argthree))
    ((numberp anyp sregp)
                         (*move argtwo (reg s6))
                         (mov (vregname argone) (reg s6) argthree))
    ((numberp anyp anyp)
                         (*move argtwo (reg S6))
                         (mov (vregname argone) (reg S6) (reg s7))
                         (*move (reg s7) argthree))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (Vset (vreg 3) x)
% set all elements of v3 to value x
%----------------------------------------------------------------------%
 
(put 'Vset  'openfn '(vector2pat *vset))
 
(de *vset (a1 a2) (expand2operandcmacro a1 a2 '*vset))
 
(defcmacro *vset
    ((numberp anysregbuts7p)
                         (*move (wconst -1) (reg s7))
                         (*VsetVm (reg s7) (reg s7))
                         (mask!.t (vregname argone) argtwo (reg V0)) )
    ((numberp anyp)
                         (*move argtwo (reg s6))
                         (*vset argone (reg s6)))
    (                    (*cerror "weird *vset")))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  INTEGER arithmetic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%----------------------------------------------------------------------%
% (vpluss (vreg 3) (vreg 5) x)
% vector summ: v3 = v5 +  scalar x
%----------------------------------------------------------------------%
 
(put 'vpluss 'openfn '(vector3pat *vpluss))
 
(de *vpluss (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vpluss))
 
(defcmacro *vpluss
    ((numberp numberp sregp)
                (VaddS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (VaddS (vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VaddS (vregname argone)(reg s6)(vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VaddS (vregname argone)(reg s6)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vplusV (vreg 3) (vreg 5) (vreg 7))
% vector summ: v3 = v5 + v7
%----------------------------------------------------------------------%
 
(put 'vplusv 'openfn '(vector3pat *vplusv))
 
(de *vplusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vplusv))
 
(defcmacro *vplusv
    ((numberp numberp numberp)
       (VaddV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (vsminusv (vreg 3) x (vreg 5))
% vector difference: v3 = scalar x - V5
%----------------------------------------------------------------------%
 
(put 'vSminusV 'openfn '(vector3pat *sminusv))
 
(de *sminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*sminusv))
 
(defcmacro *sminusv
    ((numberp sregp numberp)
                (VdiffS (vregname argone) argtwo(vregname argthree)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VdiffS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (minusv (vreg 3) (vreg 5))
% calculate : v3 =  - V5
%----------------------------------------------------------------------%
 
(put 'minusV 'openfn '(vector2pat *minusv))
 
(de *minusv (a1 a2) (expand2operandcmacro a1 a2  '*minusv))
 
(defcmacro *minusv
    ((numberp numberp)
                (VdiffS (vregname argone) (Reg S0)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VminusV (vreg 3) (vreg 5) (vreg 7))
% vector difference: v3 = v5 - v7
%----------------------------------------------------------------------%
 
(put 'vminusv 'openfn '(vector3pat *vminusv))
 
(de *vminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vminusv))
 
(defcmacro *vminusv
    ((numberp numberp numberp)
      (VminusV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  REAL arithmetic (float)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
%----------------------------------------------------------------------%
% (vfpluss (vreg 3) (vreg 5) x)
% vector summ: v3 = v5 +  scalar x
%----------------------------------------------------------------------%
 
(put 'vfpluss 'openfn '(vector3pat *vfpluss))
 
(de *vfpluss (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vfpluss))
 
(defcmacro *vfpluss
    ((numberp numberp sregp)
                (vfaddS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (vfaddS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VFaddS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argwo (reg s6))
                (VFaddS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VFplusV (vreg 3) (vreg 5) (vreg 7))
% vector summ: v3 = v5 + v7
%----------------------------------------------------------------------%
 
(put 'VFplusv 'openfn '(vector3pat *VFplusv))
 
(de *VFplusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFplusv))
 
(defcmacro *VFplusv
    ((numberp numberp numberp)
       (VFaddV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (VFsminusv (vreg 3) x (vreg 5))
% vector difference: v3 = scalar x - V5
%----------------------------------------------------------------------%
 
(put 'VFSminusV 'openfn '(vector3pat *VFsminusv))
 
(de *VFsminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFsminusv))
 
(defcmacro *VFsminusv
    ((numberp sregp numberp)
                (VFdiffS (vregname argone) argtwo(vregname argthree)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VFdiffS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (Fminusv (vreg 3) (vreg 5))
% calculate : v3 =  - V5
%----------------------------------------------------------------------%
 
(put 'FminusV 'openfn '(vector2pat *Fminusv))
 
(de *Fminusv (a1 a2) (expand2operandcmacro a1 a2  '*Fminusv))
 
(defcmacro *Fminusv
    ((numberp numberp)
                (VFdiffS (vregname argone) (Reg S0)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VFminusV (vreg 3) (vreg 5) (vreg 7))
% vector difference: v3 = v5 - v7
%----------------------------------------------------------------------%
 
(put 'VFminusv 'openfn '(vector3pat *VFminusv))
 
(de *VFminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFminusv))
 
(defcmacro *VFminusv
    ((numberp numberp numberp)
      (VFminusV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vftimess (vreg 3) (vreg 5) x)
% vector summ: v3 = v5 +  scalar x
%----------------------------------------------------------------------%
 
(put 'vftimess 'openfn '(vector3pat *vftimess))
 
(de *vftimess (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vftimess))
 
(defcmacro *vftimess
    ((numberp numberp sregp)
                (vftimesS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (vftimesS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VFtimesS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VFtimesS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (VFtimesV (vreg 3) (vreg 5) (vreg 7))
% vector product: v3 = v5 * v7
%----------------------------------------------------------------------%
 
(put 'VFtimesv 'openfn '(vector3pat *VFtimesv))
 
(de *VFtimesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFtimesv))
 
(defcmacro *VFtimesv
    ((numberp numberp numberp)
       (VFtimesV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (vfhtimess (vreg 3) (vreg 5) x)
% vector half times: v3 = v5 *  scalar x
%----------------------------------------------------------------------%
 
(put 'vfhtimess 'openfn '(vector3pat *vfhtimess))
 
(de *vfhtimess (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vfhtimess))
 
(defcmacro *vfhtimess
    ((numberp numberp sregp)
               (vfhtimesS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
               (vfhtimesS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
               (*move argthree (reg s6))
               (VFhtimesS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
               (*move argtwo (reg s6))
               (VFhtimesS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (VFhtimesV (vreg 3) (vreg 5) (vreg 7))
% vector half product: v3 = v5 * v7
%----------------------------------------------------------------------%
 
(put 'VFhtimesv 'openfn '(vector3pat *VFhtimesv))
 
(de *VFhtimesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFhtimesv))
 
(defcmacro *VFhtimesv
    ((numberp numberp numberp)
       (VFhtimesV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vrtimess (vreg 3) (vreg 5) x)
% vector prod rounded: v3 = v5 *  scalar x
%----------------------------------------------------------------------%
 
(put 'vrtimess 'openfn '(vector3pat *vrtimess))
 
(de *vrtimess (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vrtimess))
 
(defcmacro *vrtimess
    ((numberp numberp sregp)
                (vrtimesS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (vrtimesS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (vrtimesS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (vrtimesS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (vrtimesV (vreg 3) (vreg 5) (vreg 7))
% vector rounded product: v3 = v5 * v7
%----------------------------------------------------------------------%
 
(put 'vrtimesv 'openfn '(vector3pat *vrtimesv))
 
(de *vrtimesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vrtimesv))
 
(defcmacro *vrtimesv
    ((numberp numberp numberp)
       (vrtimesV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VFrezappV (vreg 3) (vreg 5))
% vector reziprocal approximantion: v3 = rez (v5)
%----------------------------------------------------------------------%
 
(put 'VFrezappv 'openfn '(vector2pat *VFrezappv))
 
(de *VFrezappv (a1 a2) (expand2operandcmacro a1 a2 '*VFrezappv))
 
(defcmacro *VFrezappv
    ((numberp numberp numberp)
       (VFrezappV (vregname argone) (vregname argtwo)                   ))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VpopcV (vreg 3) (vreg 5))
% vector popuplation count : v3 = p (v5)
%----------------------------------------------------------------------%
 
(put 'VFpopcv 'openfn '(vector2pat *VFpopcv))
 
(de *VFpopcv (a1 a2) (expand2operandcmacro a1 a2 '*VFpopcv))
 
(defcmacro *VFpopcv
    ((numberp numberp)
       (VFpopcV (vregname argone) (vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (VpopcpV (vreg 3) (vreg 5))
% vector popuplation count parity: v3 = q (v5)
%----------------------------------------------------------------------%
 
(put 'VFpopcpv 'openfn '(vector2pat *VFpopcpv))
 
(de *VFpopcpv (a1 a2) (expand2operandcmacro a1 a2 '*VFpopcpv))
 
(defcmacro *VFpopcpv
    ((numberp numberp)
       (VFpopcpV (vregname argone) (vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (vfrecits (vreg 3) x (vreg 5))
% vector reciprocal iteration: v3 = scalar x *I v5
%----------------------------------------------------------------------%
 
(put 'vfrecits 'openfn '(vector3pat *vfrecits))
 
(de *vfrecits (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vfrecits))
 
(defcmacro *vfrecits
    ((numberp numberp sregp)
                (vfrecitS (vregname argone) argthree (vregname argtwo)))
    ((numberp sregp numberp)
                (vfrecitS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VFrecitS(vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VFrecitS(vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr vfrecits" )))
 
 
%----------------------------------------------------------------------%
% (VFrecitV (vreg 3) (vreg 5) (vreg 7))
% vector reciprocal iteration: v3 = v5 *I v7
%----------------------------------------------------------------------%
 
(put 'VFrecitv 'openfn '(vector3pat *VFrecitv))
 
(de *VFrecitv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFrecitv))
 
(defcmacro *VFrecitv
    ((numberp numberp numberp)
       (VFrecitV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector instr VFrecitV")))
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  LOGICAL operations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%----------------------------------------------------------------------%
% (VandS (vreg 3) x (vreg 7))
% vector and: v3 = x & v7
%----------------------------------------------------------------------%
 
(put 'VandS 'openfn '(vector3pat *VandS))
 
(de *VandS (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VandS))
 
(defcmacro *VandS
    ((numberp sregp numberp)
      (VandS (vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
      (*Move argtwo (reg S6))
      (VandS (vregname argone) (Reg S6) (vregname argthree)))
    (                    (*cerror "weird vector &")))
 
%----------------------------------------------------------------------%
% (VandV (vreg 3) (vreg 5) (vreg 7))
% vector and: v3 = v5 & v7
%----------------------------------------------------------------------%
 
(put 'vandv 'openfn '(vector3pat *vandv))
 
(de *vandv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vandv))
 
(defcmacro *vandv
    ((numberp numberp numberp)
      (VandV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector !")))
 
%----------------------------------------------------------------------%
% (VorS (vreg 3) x (vreg 7))
% vector ohr: v3 = x ! v7
%----------------------------------------------------------------------%
 
(put 'VorS 'openfn '(vector3pat *VorS))
 
(de *VorS (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VorS))
 
(defcmacro *VorS
    ((numberp sregp numberp)
      (VorS (vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
      (*Move argtwo (reg S6))
      (VorS (vregname argone) (Reg S6) (vregname argthree)))
    (                    (*cerror "weird vector &")))
 
%----------------------------------------------------------------------%
% (VmoveV (vreg 3)  (vreg 7))
% move V7 to V3
% special case of vector ohr
%----------------------------------------------------------------------%
 
(put 'VmoveV 'openfn '(vector2pat *VmoveV))
 
(de *VmoveV (a1 a3) (expand3operandcmacro a1 '(Reg S0) a3 '*VorS))
 
(defCmacro *VmoveV)
 
 
%----------------------------------------------------------------------%
% (VorV (vreg 3) (vreg 5) (vreg 7))
% vector or: v3 = v5 ! v7
%----------------------------------------------------------------------%
 
(put 'vorv 'openfn '(vector3pat *vorv))
 
(de *vorv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vorv))
 
(defcmacro *vorv
    ((numberp numberp numberp)
      (VorV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector &")))
 
 
%----------------------------------------------------------------------%
% (VxorV (vreg 3) (vreg 5) (vreg 7))
% vector xor: v3 = v5 \ v7
%----------------------------------------------------------------------%
 
(put 'vxorv 'openfn '(vector3pat *vxorv))
 
(de *vxorv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vxorv))
 
(defcmacro *vxorv
    ((numberp numberp numberp)
      (VxorV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector \")))
 
 
%----------------------------------------------------------------------%
% special form of VxorV:
% (Vclear (vreg 3))
% clear vector register
%----------------------------------------------------------------------%
 
(put 'vclear 'openfn '(vector1pat *vclear))
 
(de *vclear (a1) (expand3operandcmacro a1 a1 a1 '*vxorv))
 
(defcmacro *vclear)
 
%----------------------------------------------------------------------%
% (vlminuss (vreg 3) (vreg 5) x)
% vector logical difference: v3 = v5 minus scalar x
%----------------------------------------------------------------------%
 
(put 'vlminuss 'openfn '(vector3pat *vlminuss))
 
(de *vlminuss (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vlminuss))
 
(defcmacro *vlminuss
    ((numberp numberp sregp)
                (VldiffS (vregname argone) argthree (vregname argtwo)))
    ((numberp sregp numberp)
                (VldiffS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VldiffS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VldiffS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (vmergevv i j k)
% vector merge: vi = vj (if vm=1) or vk
%----------------------------------------------------------------------%
 
(put 'vmergevv 'openfn '(vector3pat *vmergevv))
 
(de *vmergevv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vmergevv))
 
(defcmacro *vmergevv
    ((numberp numberp numberp)
      (Vmergevv(vregname argone) (vregname argtwo) (vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
%----------------------------------------------------------------------%
% (vmergesv i j k)
% vector merge: vi = sj (if vm=1) or vk
%----------------------------------------------------------------------%
 
(put 'vmergesv 'openfn '(vector3pat *vmergesv))
 
(de *vmergesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vmergesv))
 
(defcmacro *vmergesv
    ((numberp sregp numberp)
         (Vmergesv(vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
         (*move argtwo (reg s6))
         (Vmergesv(vregname argone) (Reg S6) (vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
 
%----------------------------------------------------------------------%
% (VlshiftV (vreg 3) (vreg 5) 17)
% vector lshift: v3 = v5 < 17
%----------------------------------------------------------------------%
 
(put 'vlshiftv 'openfn '(vector3pat *vlshiftv))
 
(de *vlshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vlshiftv))
 
(defcmacro *vlshiftv
    ((numberp numberp onep)
      (VlshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VlshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VlshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector <")))
 
%----------------------------------------------------------------------%
% (VrshiftV (vreg 3) (vreg 5) 17)
% vector rshift: v3 = v5 > 17
%----------------------------------------------------------------------%
 
(put 'vrshiftv 'openfn '(vector3pat *vrshiftv))
 
(de *vrshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vrshiftv))
 
(defcmacro *vrshiftv
    ((numberp numberp onep)
      (VrshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VrshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VrshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector >")))
 
%----------------------------------------------------------------------%
% (VdlshiftV (vreg 3) (vreg 5) 17)
% vector dlshift: v3 = v5 << 17
%----------------------------------------------------------------------%
 
(put 'vdlshiftv 'openfn '(vector3pat *vdlshiftv))
 
(de *vdlshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vdlshiftv))
 
(defcmacro *vdlshiftv
    ((numberp numberp onep)
      (VdlshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VdlshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VdlshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector <")))
 
%----------------------------------------------------------------------%
% (VdrshiftV (vreg 3) (vreg 5) 17)
% vector drshift: v3 = v5 >> 17
%----------------------------------------------------------------------%
 
(put 'vdrshiftv 'openfn '(vector3pat *vdrshiftv))
 
(de *vdrshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vdrshiftv))
 
(defcmacro *vdrshiftv
    ((numberp numberp onep)
      (VdrshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VdrshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VdrshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector >")))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% administration of vectors
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%----------------------------------------------------------------------%
% (vsetvl x)
% set vector length register
%----------------------------------------------------------------------%
 
(put 'vsetvl  'openfn '(vector1pat *vsetvl))
 
(de *vsetvl  (a1) (expand1operandcmacro a1 '*vsetvl))
 
(defcmacro *vsetvl
    ((aregp)
                (VLmoveA (reg a0) argone)) % a0 is dummy to mark argone
                                           % as source of move(lapopt)
 
    ((anyp)
                (*move argone (reg a6))
                (VLmoveA (reg a0) (reg a6)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vsetvm x)
% set vector mask register
%----------------------------------------------------------------------%
 
(put 'vsetvm  'openfn '(vector1pat *vsetvm))
 
(de *vsetvm  (a1 a2) (expand2operandcmacro a1 a2 '*vsetvm))
 
(defcmacro *vsetvm
    ((Sregp sregp)
                (mov argone argtwo (reg vm))
    ((anyp sregp)
                (*move argone (reg S6))
                (*vsetvm (reg s6) argtwo))
    ((sregp anyp)
                (*move argtwo (reg S7))
                (*vsetvm (reg s7) (reg s7)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (vsetvlsave vreg x)
% set vector length register to x and save for specific vreg
%----------------------------------------------------------------------%
 
(put 'vsetvlsave  'openfn '(vector2pat *vsetvlsave))
 
(de *vsetvlsave  (a1 a2) (expand2operandcmacro a1 a2 '*vsetvlsave))
 
(defcmacro *vsetvlsave
    ((anyp aregp)
                (VLmoveA (reg a0) argtwo)
                (move (vlthname argone) argtwo)
                (*move argtwo (reg B21))  )    % sim vl
    (%(anyp anyp)
                (*move argtwo (reg a6))
                (VLmoveA (reg a0) (reg a6))
                (move (vlthname argone) (reg a6))
                (*move (reg a6)(reg b21))  )  )
%----------------------------------------------------------------------%
% savevl / restorevl
%----------------------------------------------------------------------%
(put 'Vsavevl 'opencode '((*move (reg b21)(reg a6))
                          (*move (reg a6)(reg b22))))
(put 'Vrestorevl 'opencode '(
                         (*move (reg b22)(reg a6))
                         (*move (reg a6)(reg b21))
                         (VLmoveA (reg a0) (reg a6)) ))
%----------------------------------------------------------------------%
% read vector mask register
%----------------------------------------------------------------------%
 
%----------------------------------------------------------------------%
% (VtestZ (vreg 3) )
% test  zeroes in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestz  'openfn '(vector1valuepat *vtestz))
 
(de *vtestz  (a1 a2) (expand2operandcmacro a1  a2 '*vtestz))
 
(defcmacro *vtestz
    ((numberp sregp)
                (VMtestZ (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestZ (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VtestN (vreg 3) )
% test  NonZeroes in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestn  'openfn '(vector1valuepat *vtestn))
 
(de *vtestn  (a1 a2) (expand2operandcmacro a1  a2 '*vtestn))
 
(defcmacro *vtestn
    ((numberp sregp)
                (VMtestN (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestN (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VtestPos (vreg 3) )
% test  positives in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestpos  'openfn '(vector1valuepat *vtestpos))
 
(de *vtestpos  (a1 a2) (expand2operandcmacro a1  a2 '*vtestpos))
 
(defcmacro *vtestpos
    ((numberp sregp)
                (VMtestPos (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestPos (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
     )
 
%----------------------------------------------------------------------%
% (VtestNeg (vreg 3) )
% test  negatives in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestneg  'openfn '(vector1valuepat *vtestneg))
 
(de *vtestneg  (a1 a2) (expand2operandcmacro a1  a2 '*vtestneg))
 
(defcmacro *vtestneg
    ((numberp sregp)
                (VMtestNeg (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestNeg (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
     )
 
%----------------------------------------------------------------------%
% (Vldzero x)
% calculate leading zeroes in value x
%----------------------------------------------------------------------%
 
(put 'vldzero 'openfn '(vector1valuepat *vldzero))
 
(de *vldzero (a1 a2) (expand2operandcmacro a1  a2 '*vldzero))
 
(defcmacro *vldzero
    ((sregp anyp)
                (LDzero (reg a6) argone)
                (*move (reg a6) argtwo) )
    ((anyp anyp)
                (*move argone (reg s6))
                (LDzero (reg a6) (reg s6))
                (*move (reg a6) argtwo))
    (                    (*cerror "weird vector instr")))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  magic numbers for arithmetic
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(put 'magicSconst06 'opencode '((sconst06 (reg 1))))
(put 'magicSconst04 'opencode '((sconst04 (reg 1))))
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  definition of high level vector macros
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
% allow the writing of (vreg 3) for a vector register
 
(ds vreg (n) n)
 
% allow the writing (Vreg3) too
 
(ds Vreg0 () 0)
(ds Vreg1 () 1)
(ds Vreg2 () 2)
(ds Vreg3 () 3)
(ds Vreg4 () 4)
(ds Vreg5 () 5)
(ds Vreg6 () 6)
(ds Vreg7 () 7)
 
% VsearchVeqS:
%
% find first eq-occurence of a scalar in a vector register
% v object vector register
% s scalar value
% vaux1: auxiliary vector register
 
(ds VsearchVeqS (v s vaux1)
   (prog(u)
     (vSminusv vaux1 s v)       % vaux = s - v  arithmetic difference
     (setq u (vtestz vaux1))   % test vaux to zero
     (return
         (cond ((eq u 0) -1)   % not found: -1
               (t (vldzero u)) % found: calculate position number
))))
