%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  FILE:       COMPILER
%
%  Description:  load PSL 3.4 Compiler for CRAY
%
%  Author:     Winfried Neun, ZIB Berlin
%  Date  :     March 1988
%
%  Copyright (c) 1988, Konrad Zuse-Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% load the compiler; modify it for Y-MP (X-MP EA resp)
 
(load1      'datamachine)
(load1      'pass-1 )
(load1      'pass-2-3)
(load1      'pass-1-cray )
(load1      'pass-3-cray )
(load1      'comp-decls )
(load1      'pass-1-lap )
(load1      'cray-lap )
(load1      'cray-forms )
(load1      'cray-instrs )
(load1      'cray-cmac )
(load1      'cray-comp )
(load1      'faslout)
(load1      'huge-frame)
 
(compiletime (load if-system))

(if_system Y-MP       % rest is for Y only
(progn
(compiletime (setq *syslisp t)) % for symval etc.
 
(compiletime(progn
 
(put 'reloc_word          'const 1)
(put 'reloc_id_number     'const 2)
(put 'escape_to_second    'const 3)
(put 'parcel_label        'const 0)
(put 'word_label          'const 1)
(put 'reloc_value_cell    'const 2)
(put 'reloc_function_cell 'const 3)
nil))
 
% Following instructions are changed:
 
%Ai   exp   020,021
%Si   exp   040,041
%Ai exp,Ah  10h
%exp,Ah Ai  11h           3-parcel intructions
%Si exp,Ah  12h
%exp,ah Si  13h
 
%No more Ema instruction
 
%Vi Sj*IVk   is now Vi Sj*Vk (166) no more Imul but int multi
 
%The  model for fasl etc:
% Three parcel instructions are coded in bittable as before the
% correspondind  2 parcel instructions + one extra 0 parcelbit
% Problem: PLOC must differentiate between jumpop and Ai  exp
 
%CRFORMS 
 
(de y-mpart (e)    % m field of instruction
    (wand  16#FFFF (wrshift e 16)))
 
(de y-npart (e)    % n field of instruction
    (wand e 16#FFFF))
 
(de op-i-mn (op i mn)
    (setq mn (addressvalue mn nil))
    (threeparcel (wor (opvalue op) (wlshift i 6))
                 (y-mpart mn) (y-npart mn)))
 
(de op-h-i-mn (op h i mn)
   (setq mn (Parceladdressvalue mn nil))
   (threeparcel (wor (wlshift op 12) (wor (wlshift h 9) (wlshift i 6)))
                 (y-mpart mn) (y-npart mn)))
 
(de ploc-addressvalue (e *test-only*)  % spec for PLOC instruction
   (if (labelp e)                      % must be different from Jump
       (progn
        (when (and *writingfaslfile (not *test-only*))
              (add-next-bit (const escape_to_second))
              (add-next-bit (const parcel_label))
              (add-next-bit 0))
        (labelvalue e))
       (print (list e *test-only*))
       (error 999 "weird ploc-addressvalue")))
 
(de addressvalue (e *test-only*)
   (cond ((minusexpp e) e)
         ((posintp e) e)
         ((labelp e) (cond (*writingfaslfile
                       (cond ((not *test-only*)
                            (add-next-bit 0)
                            (add-next-bit (const escape_to_second))
                            (add-next-bit (const word_label))))
                       (parceltoword (labelvalue e)))
                     (t (parceltoword (labelvalue e)))))
         ((not (regp e)) (operandvalue e))))
 
% The location of the value cell.
(de valuecelllocation (u)
  (cond ((not *writingfaslfile)
         (wplus2 symval (id2int u)))
        (t (cond ((not *test-only*)
                (add-next-bit 0)
                (add-next-bit (const escape_to_second))
                (add-next-bit (const reloc_value_cell))))
           (findidnumber u))))
 
(de ExtraRegLocation (X)
  (progn 
   (setq X (second X))
   (cond ((not *WritingFaslFile)
            (wplus2 ArgumentBlock
                (wdifference X (wplus2 MaxRealRegs 1))))
         (t (cond ((not *test-only*)
                    (add-next-bit 0)
                    (add-next-bit (const escape_to_second))
                    (add-next-bit (const reloc_value_cell))))
            (wplus2 X 8150))))  )
 
 
(de Threeparcel (x y z)
   (prog (Newaddr Outputbase*)
        (setq outputbase* 8)
        (setq Newaddr (calculateaddress))
        (when *pcode (printf "%p%t%pp%w%t%p%t%p%t%p%n" currentoffset*
                         10 (wrshift newaddr 2) (parcelletter Newaddr)
                         25 x 40 z 55 y))
        (Crayputparcel newaddr x)
        (Crayputparcel (wplus2 newaddr 1) z) % reversed
        (Crayputparcel (wplus2 newaddr 2) y)
        (setq Currentoffset* (wplus2 Currentoffset* 3))))
 
% Crinstr
 
   % load of relocateable parceladdress into a-register
 
(defOpCode PLOC (r1 r2)
  ((and (aregp r1) (ploc-addressvalue r2 t))
        (Op-i-mn 8#020 (aregp r1) (ploc-addressvalue r2 nil))))
 
(put      'PLOC 'LapLength 3)
 
% %%%%%%%%%%%%%  Loads and Stores %%%%%%%%%%%%%%%%%%%%%%%
 
(de storeOpCodeFexpr (AllArgs)
 (prog (r1 e a2)
   (setq r1 (first AllArgs))
   (cond ((equal (length AllArgs) 2)
               (cond ((eqcar (second AllArgs) 'Indexed)
                        (setq a2 (second (second AllArgs)))
                        (setq e (third (second AllArgs))))
                     (T (setq a2 '(reg a0))
                        (setq e (second AllArgs)))))
         (T (setq e (second AllArgs))
                (setq a2 (third AllArgs))))
 (return 
  (cond
    %      e,a2   r1
    ((and (AregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-mn 8#11 (AregP a2) (AregP r1) (AddressValue e nil)))
    ((and (SregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-mn 8#13 (AregP a2) (SregP r1) (AddressValue e nil)))
    (T (OpFailed 'Store))))))
 
(put 'STORE 'LapLength 3)
 
(de LoadOpCodeFexpr (AllArgs)
  (prog (r1 e a2)
    (setq r1 (first AllArgs))
    (cond ((equal (length AllArgs) 2)
           (cond ((eqcar (second AllArgs) 'Indexed)
                       (setq a2 (second (second AllArgs)))
                       (setq e (third (second AllArgs))))
                 (T (setq a2 '(reg a0))
                    (setq e (second AllArgs)))))
         (T (setq e (second AllArgs))
            (setq a2 (third AllArgs))))
 (return 
  (cond
    %      r1    e,a2
    ((and (AregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-mn 8#10 (AregP a2) (AregP r1) (AddressValue e nil)))
    ((and (SregP r1) (AregP a2) (AddressValue e T))
      (Op-h-i-mn 8#12 (AregP a2) (SregP r1) (AddressValue e nil)))
    (T (OpFailed 'Load))))))
 
(put 'LOAD 'LapLength 3)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% MOVE %%%%%%%%%%%%%%%%%%%%%%%%%
% Register r2 to register r1, or const
 
(defOpCode      MOVE    (r1 r2)
%      r1      r2
 ((and (Aregp r1) (JKSizeP r2))
                (Op-i-jk 8#022 (Aregp r1) r2))
 ((and (Aregp r1) (MinusExpP r2))
                (Op-i-mn 8#021 (Aregp r1) (TwosComp r2)))
 ((and (Aregp r1) (AddressValue r2 T))
                (Op-i-mn 8#020 (AregP r1) (AddressValue r2 nil)))
 ((and (Aregp r1) (Sregp R2))
         (Op-i-j-k  8#023 (AregP r1) (SregP r2) 0))
 ((and (Aregp r1) (Bregp R2))
         (Op-i-jk  8#024 (AregP r1) (BregP r2) ))
 ((and (Bregp r1) (Aregp R2))
         (Op-i-jk  8#025 (AregP r2) (BregP r1) ))
 ((and (Aregp r1) (Aregp r2))
                (Op-i-j-k 8#030 (Aregp r1) 0 (Aregp r2)))
 ((and (Sregp r1) (Onep r2))
                (Op-i-jk 8#042 (Sregp r1) 8#77))
 ((and (Sregp r1) (MinusOneP r2))
                (Op-i-jk 8#042 (Sregp r1) 8#00))
 ((and (Sregp r1) (Zerop r2))
                (Op-i-jk 8#043 (Sregp r1) 8#00))
 ((and (Sregp r1) (MinusExpP r2))
                (Op-i-mn 8#041 (Sregp r1) (TwosComp r2)))
 ((and (Sregp r1) (AddressValue r2 T))
                (Op-i-mn 8#040 (Sregp r1) (AddressValue r2 nil)))
 ((and (Sregp r1) (Sregp r2))
                (Op-i-j-k 8#051 (Sregp r1) 0 (Sregp r2)))
 ((and (Sregp r1) (Aregp r2))
                (Op-i-j-k 8#071 (Sregp r1) 0 (Aregp r2)))
 ((and (Sregp r1) (Tregp r2))
                (Op-i-jk 8#074 (Sregp r1) (Tregp r2)))
 ((and (Tregp r1) (Sregp r2))
                (Op-i-jk 8#075 (Sregp r2) (Tregp r1)))
)
 
(put 'MOVE 'LapLength 'LapMoveLength)
 
% End of MOVE definition table
 
(de LapMoveLength (U)
 (prog (r1 r2)
  (setq r1 (second U))
  (setq r2 (third U))
  (return
   (cond 
     ((and (Aregp r1) (JKSizeP r2))        1)
     ((and (Aregp r1) (MinusExpP r2))      3)
     ((and (Aregp r1) (ParcelP r2))        3)
     ((and (Aregp r1) (AddressValue r2 T)) 3)
     ((and (Aregp r1) (Sregp R2))          1)
     ((and (Aregp r1) (Bregp R2))          1)
     ((and (Bregp r1) (Aregp R2))          1)
     ((and (Aregp r1) (Aregp R2))          1)
     ((and (Sregp r1) (Onep r2))           1)
     ((and (Sregp r1) (MinusOneP r2))      1)
     ((and (Sregp r1) (Zerop r2))          1)
%     ((and (Sregp r1) (MinusExpP r2))     3)
     ((and (Sregp r1) (ParcelP r2))        3)
     ((and (Sregp r1) (AddressValue r2 T)) 3)
     ((and (Sregp r1) (Sregp r2))          1)
     ((and (Sregp r1) (Aregp r2))          1)
     ((and (Sregp r1) (Tregp r2))          1)
     ((and (Tregp r1) (Sregp r2))          1)
     (T (printf "Can't find length of case: %p%n" U)))
 )))
 
% Cmacs1, Literal size etc
 
(de SmallInumP (Expression)
 (or (and (eqcar expression 'immediate)
          (eqcar (cadr expression) 'xidloc))
     (and (InumP Expression)
          (wleq Expression         16#7FFFFFFF)
          (wgeq Expression (wminus 16#40000000)))))
 
(de binxidloc (u)
  (cond ((not *writingfaslfile)
            (wplus2 (wtimes2 (eval (list 'idloc (cadr u)))
                        &LinkShift&) (caddr u)))
        (t (cond ((not *test-only*)
              (add-next-bit 0)   % Reloc ID, next says to do xidloc.
              (add-next-bit 2)   % (const reloc_id_number))))
              (add-next-bit 1))) % (const reloc_id_number))))
           (wplus2 (wtimes2 (findidnumber (cadr u))
                           &LinkShift&) (caddr u)))))
 
(compiletime (setq *syslisp nil)) % for symval etc.
))
