%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CRAY-CMAC-1.SL -  CRAY PSL cmacros, Part 1                            %
%                                                                       %
% Author:      M. L. Griss, S. Lowder, R. Kessler                       %
% (PSL3.2)     Computer Science Dept.                                   %
%              University of Utah                                       %
% Date:        1 June 1982                                              %
%                                                                       %
% Based on M68K-CMAC.SL  1 June by Lowder et al                         %
%                                                                       %
% Revisions    changed nearly all for Cray PSL 3.4                      %
%              by H. Melenk and W. Neun,ZIB Berlin                      %
%                                                                       %
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Modified (nearly all) by ZIB Berlin starting January 1986
%  for various reasons:
%              migration from VAX/VMS  to CRAY/COS
%              migration from PSL 3.2 to PSL 3.4
%  This is the PSL 3.4 Version
%
%  <PSL.COMP.CRAY>CRAY-CMAC.SL.5,  4-Aug-83 16:02:37, Edit by KESSLER
%  Flag Indexed as terminaloperand.  Problem was discovered on the Vax in
%  which Indexed is not flagged as terminaloperand, but is regularly done
%  on the 20.  Also fixed a bug in Martin's JUMPON of WDIFFERENCE.
%  <PSL.COMP.CRAY>CRAY-CMAC.SL.3,  1-Aug-83 10:42:15, Edit by GRISS
%  Bug in JUMPON not handling nonzero lower bound properly.
 
(flag '(Immediate Indexed UnImmediate QuotedId) 'TerminalOperand)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MACHINE DEPENDENT CONSTANTS FOR THE CRAY                        %
% Stack grows up, rather than down, used to control Stack Scan loops.   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 %  tag | reloc | inf
 %   5b    22b     37b
 % 0     5      27
 
(setq *ImmediateQuote NIL)               % Force A LIT word for Item
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Fetch the Predicates for the following procedures.                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% 5/9/85 - Kessler - Added Predicates for stack frame in T register stuff.
% 5/9/85 - Kessler - Added T3 -> t7 to T register list.
%  Added T as a T register.  T will now be kept in a register.  This
%  May not be necessary, since NIL isn't flagged as a T register.
% 5/8/85 - Kessler - Added Stack and Heap maxies as TREGS.
% 12/27/84 - Kessler
%  New version of smallinump.
%  <PSL.COMP.CRAY>PREDICATES.SL.26,  8-Aug-83 14:13:42, Edit by KESSLER
%  Added FakeRegisterNumberP stolen from the 20.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%             REGISTER HANDLING PREDICATES                              %
% These handle specific predicate cases                                 %
% for the CRAY    not supported by the standard predicates outlined in  %
% the CMACRO guide. The common predicates can be used in these CMAC's;  %
% they are defined the the file PC:Common-Predicates.SL                 %
%                                                                       %
% InumP tells what numbers can be immediate operands on the target      %
%      machine.  E.g., specifies they must be FixPoint and with in      %
%      valid range for a 37 bit integer.                                %
%                                                                       %
% SmallInumP is a subcase of InumP which checks for integers not greater%
%      than 22 bits.                                                    %
%                                                                       %
% PosInumP is a subcase of InumP excluding negatives.                   %
%                                                                       %
% NegInumP is the same as InumP except it tests for negative Inum`s.    %
%                                                                       %
% PosShiftInumP tests for shifts of 1 to 64.                            %
%                                                                       %
% NegShiftInumP      test for shifts of -1 to -64.                      %
%                                                                       %
% the following register tests are in Cray forms
%
% RegP tests if the operand is a valid CRAY register.                   %
%                                                                       %
% AregP tests for a Cray A or address register, 24 bits.                %
% BregP tests for a Cray B register which is an A register cache.       %
%                                                                       %
% SregP tests for a Cray S for a data register, 64 bits.                %
% TregP tests for a Cray T register which is an S register cache.       %
% VregP tests for a Cray vector register                                %
%                                                                       %
% ImmediateP tests if an item is tagged IMMEDIATE.                      %
%      WConsts and WArrays are tagged immediate when they are not       %
%      inside MEMORY. The tagging means that the following expression   %
%      is to be used as an immediate value. For example, if WArray      %
%      SYMFNC is the base of this table, the expression                 %
%      (*WPLUS2 (Reg 1)(WArray SYMFNC)) will have the (WArray SYMFNC)  %
%      tagged immediate meaning to add the address of SYMFNC to (Reg 1) %
%      and not the contents of the SYMFNC location. An immediate        %
%      expression example would be (*Move (autoincrement (Reg st))     %
%      (plus2 (WArray ArgumentBlock) 32)). The destination expression   %
%      would be prefixed with the tag immediate.                        %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de TaggedLabel (X)
  (or (StringP X) (EqCar X 'Label)))
 
(DE InumP(EXP)
     (Or (Eq (setq Exp (Wshift Exp -37)) 0)
         (Eq  Exp 16#7FFFFFF)
)    )
 
(de SmallInumP (Expression)
 (or (and (eqcar expression 'immediate)
          (eqcar (cadr expression) 'xidloc))
     (and (InumP Expression)
          (leq Expression 16#3fffff)
          (geq Expression (minus 16#400000)))))
 
(de PosInumP (Expression)
  (and (InumP Expression) (geq Expression 0)))
 
(de NegInumP (Expression)
  (and (InumP Expression) (minusp Expression)))
 
(de PosShiftInumP (Expression)
  (and (FixP Expression) (geq Expression 1) (leq Expression 64)))
 
(de NegShiftInumP (Expression)
  (and (FixP Expression) (geq Expression (minus 64)) (leq Expression 0)))
 
(de WordWidthP (Expression)
    (and (FixP Expression) (eqn Expression 64)))
 
(de IndexedP (E) (EqCar E 'Indexed))
 
(de CdrP (e) (and (IndexedP e) (eqcar (cddr e) 1)))
 
(de Registernamep (u) (get u 'Registername))
 
(de AnyAregButA7P (RegName)
          (AND (RegP RegName)
               (MemQ (cadr RegName) '(a0 a1 a2 a3 a4 a5 a6 st))))
 
(De AsregP (x)    (or (Aregp x) (SregP x)))
 
(De ABregP (x)    (or (Aregp x) (BregP x)))
 
(de AnySregButS7P (regName)
         (AND (RegP RegName)
              (MemQ (cadr RegName) '(S0 S1 S2 S3 S4 S5 S6 1 2 3 4 5))))
 
(de TempSregP(regName) (AND (RegP RegName) (MemQ (cadr RegName) '(S6 S7))))
 
%(de A0RegP (RegName) (AND (RegP RegName) (EQ (cadr RegName) 'a0)))
 
%(de S0regP (regName) (AND (RegP RegName) (EQ (cadr RegName) 'S0)))
 
%(de ANon0RegP (RegName)
%                (AND (RegP RegName)
%                    (MemQ (cadr RegName) '(a1 a2 a3 a4 a5 a6 a7 st))))
 
%(de SNon0regP (regName)
%              (AND (RegP RegName)
%                  (MemQ (cadr RegName) '(S1 S2 S3 S4 S5 S6 S7 1 2 3 4 5))))
 
(DEFLIST '((NIL "T00")  % This reg always holds NIL.
           (T1 "T01")   % Use the data registers for temps
           (T2 "T02")
           (T3 "T03")
           (T4 "T04")
           (T5 "T05")
           (T6 "T06")
           (T7 "T07")
 % are the following 4 fluids heavily used ??? W.Neun
           (StackLowerBound  "T11")
           (StackUpperBound  "T12")
           (HeapLowerBound   "T13")
           (HeapUpperBound   "T14")
           (&symflag&        "T15")
           (ArithArgs0       "T16") % Arguments for arithmetic
           (ArithArgs1       "T17")
           (HEAPTRAPPED      "T20")
           (T                "T21")  % T is now in a register.
           (betahi*          "T22")
           (betalow*         "T23")  % for arithmetic
           (currentscantable* "T24")
           (iobuffer         "T25")
% T26 -> T70 in use as stack frame.
% T71 -> T76 are used for fast-frame
% T26 is the first of the Stack frame in the T registers.
           (PREVFRAMESIZE    "T26")
           (RETURNADDR       "T27")
           (NARGREG          "T77")
% A-REGISTER
           (ST "A4")
% B-REGISTER                         W.Neun 28-May-86
           (nextposition     "B30")
           (lineposition     "B31")
           (writefunction    "B32")  % SOME FLUIDS ARE NOW IN B-REGS
           (readfunction     "B33")
           (maxline          "B34")
           (chtokentype      "B35")
           (currentchar      "B36")
           (bufferlength     "B37")
           (tokch            "B40")
           (symget           "B41")
 
           (unreadbuffer     "B43")
           (lambda-link-address "B44")
           (codeptr*         "B45")
           (undefined-function-address "B46")
           (catchstackptr    "B47")
           (SYMVAL           "B50")
           (SYMFNC           "B51")
           (SYMPRP           "B52")  % SOME FLUIDS ARE NOW IN B-REGS
           (SYMNAM           "B53")
           (STKUPBND         "B54")
           (HashTable        "B55")
           (BndstkPtr        "B56")
           (TokenBuffer      "B57")
           (HEAPLAST         "B60")
           (HEAPTRAPBOUND    "B61")
           (bndstkupperbound "B62")
           (argumentblock    "B63")
           (bndstklowerbound "B64")
           (codeargs         "B65")
           (symcnt           "B66")
           (lambindargs*     "B67")
           (in*              "B70")
           (out*             "B71")
           (big-preallocated* "B72")
           (*bigcompress-threshold* "B75"))
         'REGISTERNAME)
 
(de AregOperatorP (Operator) (MemQ Operator '(ADD SUB)))
 
(de AddOperatorP (Operator) (EQ Operator 'ADD))
 
(de SubOperatorP (Operator) (EQ Operator 'SUB))
 
(de ImmediateP (x)
       (or (IDP x) (and (EqCar x 'Immediate) (Null (fixp (cadr x))))))
 
(De PosLE64P (x)  (and (FixP x) (leq x 64) (geq x 0)))
 
(De NegLE64P (x)  (and (FixP x) (geq x -64) (leq x 0)))
 
(De WZeroP (x) (or (Zerop x) (equal x '(wconst 0))))
 
(De QuoteP (x)    (EQCAR x 'QUOTE))
 
(De QuotedIdP (x) (EQCAR x 'QUOTEDID))
 
(De AsmModeP (x) % Legal "escapes" in Move
  (AND (PairP x) (or (Memq (car x)
                       '(fluid $fluid global $global memory extrareg))
                     (and (eqcar x 'REG) (numberp (cadr x))
                                         (greaterp (cadr x) 5)))))
 
% used to Allocate A6 or A7 when needed
(De RegT1P(x)  (And (Eqcar x 'REG) (Eq (cadr x) 'T1)))
 
(De RegT2P(x)  (And (Eqcar x 'REG) (Eq (cadr x) 'T2)))
 
(de FakeRegisterNumberP (Expression)
         (and (fixP Expression) (GreaterP Expression 5)))
 
(de FitsInStackP (x)  (wlessp x 34))
 
(fluid '(&Links-Present&))
 
% Check to see if the argument is zero, and that we don't have any links
% in the code (which is determined in the lap function).  When this is
% true, we do not need to generate any alloc 0 or dealloc 0 code.  This
% means that any leaf functions will go very very fast^^^^^^^^
%
% If a function has no links and a small framesize, we can allocate
% the frameregisters in the free regs t71-t76. This way no frame must
% be saved on stack.
%
 
(de FreeP (alloc-size) (and (eq alloc-size 0) (not &Links-Present&)))
 
(fluid '(alloc-size*))
 
(de Fast-Frame (alloc-size)(and (wlessp (setq alloc-size* alloc-size) 7)
                                 (not &Links-Present&)))
 
(de Fast-Frame-Aux (dummy) (and (wlessp alloc-size* 7)
                                 (not &Links-Present&)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       ANYREG DEFINITIONS                              %
% These are the patterns for anyreg definitions.  They are of the       %
% form:                                                                 %
% DefAnyreg ANYREG_NAME RESOLUTION_FUNCTION NAME                        %
%            ((Register_Predicate1) (Returned value))                   %
%            ((Register_Predicate2) (returned value))                   %
%                  :                   :                                %
%                        ..etc.                                         %
% The last element in the defintion has no predicate, it is a fall-thru %
% (default) case.  "Anyregs" are Flagged (on the property list) with    %
% 'ANYREG. They have property names ANYREGPATTERNTABLE- The Predicate   %
% value list; and ANYREGRESOLUTIONFUNCTION- The function called to      %
% resolve the Anyreg.                                                   %
%                                                                       %
% The associated functions of most ANYREG's are called with 2           %
% arguments, a temporary REGISTER, used in the computation if needed,   %
% and the single SOURCE is acting as an extended addressing mode.       %
% The associated function usually uses (OneOperandAnyreg Register       %
% Source 'Anyregname) or (ExpandOneArgumentAnyReg Register Source       %
% 'AnyregName). The 'AnyregName is used to find the associated pattern. %
% The MEMORY anyreg has an additional argument, called ARGTWO. The      %
% (OneOperandAnyreg Register Source Anyregname)                         %
% calls the lower level ExpandOneArgumentAnyreg, after replacing Source %
% by (ResolveOperand Register Source).  This function processes Source  %
% to see if it is. The MEMORY anyreg means that the location is the     %
% address of the source or destination.                                 %
%      It is important that the aynregs DO NOT assign any explicit      %
% registers because an instruction may have two similar anyregs which   %
% cause the explicit register to be used twice in the same instruction. %
% The future may require a scheme to allocate temporary A and D regs.   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Crude Areg allocator, key A1,A2 A3 to T1 T2 T3
 
(de TempRegP (x)  (and (eqcar x 'Reg) (memq (cadr x) '(t1 t2 t3))))
 
(de reg (x) (list 'reg x))
 
(FLUID '(ResultingCode*))
 
(DefAnyreg CAR OURANYREGCAR)
(DefAnyreg CDR OURANYREGCDR)
 
(DE OURANYREGCAR (REG SOURCE)
   (PROG()
     (setq SOURCE (ResolveOperand REG SOURCE))
     (RETURN (OURANYREGMEMORY REG SOURCE  0))
)  )
 
(DE OURANYREGCDR (REG SOURCE)
   (PROG()
     (setq SOURCE (ResolveOperand REG SOURCE))
     (RETURN (OURANYREGMEMORY REG SOURCE  1))
)   )
 
(fluid '(&Frame-Tregs& &Fast-Frame-Tregs&))
 
(setf &Frame-Tregs&
 [ nil T30 T31 T32 T33 T34 T35 T36 T37 T40 T41 T42 T43 T44 T45 T46 T47
   T50 T51 T52 T53 T54 T55 T56 T57 T60 T61 T62 T63 T64 T65 T66 T67 T70
 ])
(setf &Fast-Frame-Tregs&  [ nil T71 T72 T73 T74 T75 T76 ])
 
(DEFANYREG FRAME FRAMEREG)
 
(DE FrameReg (Dum Number) (Cond ((Fast-Frame-Aux dum)
                             (List 'Reg (Indx &Fast-FRAME-TREGS& Number)))
                             (t (List 'Reg (Indx &FRAME-TREGS& Number)))
)                         )
 
(De AnyRegImmediate (REGISTER SOURCE)
                (cond ((InumP  SOURCE)  SOURCE)
                      ((Eqcar  SOURCE 'Unimmediate) SOURCE)
                      (T (list 'immediate SOURCE))))
 
(DefAnyReg IMMEDIATE AnyRegImmediate)
 
(DEFANYREG MEMORY OURANYREGMEMORY )
 
(FLUID '(OURAregS *AregCOUNT))
 
(setq *AregCOUNT 0)
 
(setq OurAregs '((REG A5)(REG A2)(REG A1)))
 % don't worry, but A6 and A7 is heavily used in cmacros
 
(de Areg1-3P (x) (member x OurARegs))
 
(De NextAreg () (Cond
                  ((wlessp (setq *AregCOUNT (WPlus2 *AregCOUNT 1)) 4)
                             (NTH OURAregS *AregCOUNT) )
                  ( T (NTH OURAregS (setq *AregCOUNT 1)) )
)               )
 
(De OurAnyRegMemory ( REG ARGONE ARGTWO )
   (PROG (Areg)
        (setq argone (ResolveOperand Reg argone))
        (setq argtwo (ResolveOperand Reg argtwo))
        (cond ((Aregp argone) (return `(indexed ,argone ,argtwo))))
        (setq Areg (cond ((and (eqcar argone 'indexed)
                               (AReg1-3p (cadr argone)))
                                           (cadr argone))
                          (t (nextAreg)))) %save Aregs
                                 % but never take Reg st !!!
        (setq RESULTINGCODE*
              (CONS (LIST '*Move ARGONE Areg) RESULTINGCODE*))
        (Return (list 'indexed Areg argtwo))
) )
 
(DEFANYREG $aplus2 ouraplus2 )
 
(de ouraplus2 (reg argone argtwo)
       (prog (Areg Areg2)
           (setq argone (ResolveOperand Reg argone))
           (setq argtwo (ResolveOperand Reg argtwo))
           (cond ((zerop argone) (return argtwo))
                 ((zerop argtwo) (return argone))
                 ((and (AReg1-3p argone) (Areg1-3p argtwo))
                    (setq resultingcode*
                      (cons `(*wplus2 ,argone ,argtwo) resultingcode*))
                    (return argone))
                 ((and (intp argone) (intp argtwo))
                            (return (wplus2 argone argtwo)))
                 ((and (onep argone) (Areg1-3p argtwo))
                    (setq resultingcode*
                      (append
                        `((*wplus2 ,argtwo 1))
                         resultingcode*))
                            (return argtwo))
                 ((and (onep argtwo) (Areg1-3p argone))
                    (setq resultingcode*
                      (append
                        `((*wplus2 ,argone 1))
                         resultingcode*))
                            (return argone))
                 ((onep argone)
                    (setq areg (nextareg))
                    (setq resultingcode*
                      (append
                        `((*wplus2 ,areg 1)
                          (*Move ,argtwo ,areg))
                         resultingcode*))
                            (return areg))
                 ((onep argtwo)
                    (setq areg (nextareg))
                    (setq resultingcode*
                      (append
                        `((*wplus2 ,areg 1)
                          (*Move ,argone ,areg))
                         resultingcode*))
                            (return areg))
                 ((Areg1-3p argone)
                   (setq resultingcode*
                     (append
                       `((*wplus2 ,argone (Reg A6))
                         (*Move ,argtwo (Reg A6)))
                         resultingcode*))
                    (return argone))
                 ((Areg1-3p argtwo)
                    (setq resultingcode*
                      (append
                        `((*wplus2 ,argtwo (Reg A7))
                          (*Move ,argone (Reg A7)))
                         resultingcode*))
                    (return argtwo))
                 (t (setq Areg (nextAreg))
                    (setq resultingcode*
                      (append
                        `((*wplus2 ,Areg (Reg A6))
                          (*Move ,argtwo (Reg A6))
                          (*Move ,argone ,Areg))
                         resultingcode*))
                    (return Areg))
)      )   )
 
(DEFANYREG $atimes2 ouratimes2 )
 
(de ouratimes2 (reg argone argtwo)
       (prog (Areg Areg2)
           (setq argone (ResolveOperand Reg argone))
           (setq argtwo (ResolveOperand Reg argtwo))
           (cond ((and (Areg1-3p argone) (Areg1-3p argtwo))
                    (setq resultingcode*
                      (cons `(*wtimes2 ,argone ,argtwo) resultingcode*))
                    (return argone))
                 ((Areg1-3p Argone)
                    (setq resultingcode*
                      (append
                        `((*wtimes2 ,argone (Reg A6))
                          (*Move ,argtwo (Reg A6)))
                         resultingcode*))
                    (return argone))
                 ((Areg1-3p argtwo)
                    (setq resultingcode*
                      (append
                        `((*wtimes2 ,argtwo (reg A7))
                          (*Move ,argone (reg A7)))
                         resultingcode*))
                    (return argtwo))
                 (t (setq Areg (nextAreg))
                    (setq resultingcode*
                      (append
                        `((*wtimes2 ,Areg (Reg A7))
                          (*Move ,argtwo  (Reg A7))
                          (*Move ,argone ,Areg))
                         resultingcode*))
                    (return Areg))
)      )   )
 
(DefAnyreg QUOTE
         AnyregQUOTE
         ((Null) (REG T00))  %If NIL, return the NIL register
         ((EqTP) (REG T21))  %?
         ((Idp)  (QuotedId Source)) % QuotedIds are Moved differently
         ((InumP) SOURCE)    %?
         ((QUOTE SOURCE)))   %? Recursivly expand??
 
(DefAnyreg REG  AnyregREG
         ((FakeRegisterNumberP) (ExtrAreg SOURCE))
         (                      (REG SOURCE)))
 
% change ($fluid ...) into (reg ... )  if ... has a registername
% and syslisp is turned on (for safety).
% or it is flagged *user-variable
% Treat like former WVAR's.
 
(loadtime (remflag '($fluid $global) 'terminaloperand) )
 
(de regnamep-syslisp (u) (and (registernamep u) (or *syslisp
                                (flagp u '*user-variable))))
 
(de anyregfluid (register source)
     (expandoneargumentanyreg register source '$Fluid))
 
(defanyreg $Fluid anyregfluid           % Check source  registername?)
        ((regnamep-syslisp) (reg source)) % change into (REG source)
        (($fluid source))
)
 
(de anyregglobal (register source)
     (expandoneargumentanyreg register source '$global))
 
(defanyreg $global anyregglobal         % Check source  registername?)
        ((regnamep-syslisp) (reg source)) % change into (REG source)
        (($global source))
)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  CMACRO DEFINITIONS                                   %
% Remember that it IS legal to use Other CMACROS in expansions          %
% or to define "fake" cmacros, ie procedures, for                       %
% Common cases. If a new C-macro is not standard,                       %
% e.g (*WFOO argone argtwo), then define its table and procedure.       %
% You may have to define some more "anyregs" or 'TerminalOperands       %
% Or define predicates to be CAREFUL.                                   %
% (def Cmacro *WFOO                                                     %
%      ...... )                                                         %
% (de *WFOO (Arg1 Arg2)                                                 %
%   (Expand2OperandCmacro Arg1 Arg2 '*Wfoo))                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         CMACROS.SL
% Description:  Cmacros for the Cray
% Author:       Robert R. Kessler, Utah PASS Project
% Created:      ??-Mar-83
% Modified:     08-May-85 10:53:21 (Robert R. Kessler)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1985, University of Utah, all rights reserved
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 25-May-86 - W.Neun - Added Import for opencoded functions
%                  Improved add and times for BRegs
%                  removed myanyreglabelgen and ...ref
%  3-Feb-86 - W.Neun - Added Definition of *Movesgn
% 20-May-85 - Wayne Anderson - In WQuotient, changed Move to MoveSGN in
% when the quotient is Moved from an A register to an S register
% 16-May-85 - William Galway - Fixed things so that calls on Wquotient
% function turn into invocations of *Wquotient cmacro.
% 15-May-85 - William Galway - Added *Wquotient, does special cases for
%  division by one and other powers of two, never calls external FORTRAN
%  routine.
% 13-May-85 - Kessler - ELiminated use of LinkReg, stored both in Nargreg.
%  Made it an operand, instead of a complicated expression.  Need to also
%  change the faslin stuff so the id number is handled properly.
% 09-May-85 - Kessler - Added new alloc,dealloc for Treg Stack
%  Also, changed push to generate a cerror, since it shouldn't be used.
%  Fixed bug in *loc and Tregister not present.  Also blkld -> blklod
% 09-May-85 - Kessler - Changed references to T16 to T7 (temp register).
% 08-May-85 10:03:39 (Robert R. Kessler)
%  Optimized *Tag (used 3 address code), *Jumpeq, *Jumpneq (both use
%  xor instead of -), and *jumpnottype and *jumptype (had extraneous
%  Move statement).  *link and *linke if internal functions then no
%  need to do the makelinkregs.
%
% 12/28/84 - Kessler
%  Changed the const reloc_id_number from binxidloc, so it uses the actual
%  constant.  WARNING FASLIN.SL DEFINED THE CONSTANT as a WCONST^^^^^
% 12/27/84 - Kessler
%  New version of *Move cmacro that knows about parcelp.
%  Put in labelgen, and myanyreg versions of labelgen, etc.
%  Also put in addressapply0 macros.
% 28-Nov-83 13:07:12 (Robert Kessler)
%  Made binxidloc look at the *test-only* flag to see if we are
%  putting out bits.
% 27-Nov-83 12:47:20 (Robert Kessler)
%  Changed binxidloc, so it knows about fasl files and will turn into
%  a relocatable id.
% Kessler - 3/22/84 - WTIMES2 needed to signed Move.
%  <PSL.COMP.CRAY>CMACROS.SL.7, 22-Jun-83 08:15:34, Edit by KESSLER
%  Change LMask to RMask in putfield and cmacro
%  <PSL.COMP.CRAY>CMACROS.SL.4,  4-Apr-83 07:17:23, Edit by KESSLER
%  It turns out there is no 3 operand Cmacro.  So will use 4 and ignore
% last one.
%  <PSL.COMP.CRAY>CMACROS.SL.3,  3-Apr-83 10:53:16, Edit by KESSLER
%  Changed 2 operand *Merge to its proper 3 operand class.
%  <PSL.COMP.CRAY>CMACROS.SL.12, 30-Mar-83 09:26:38, Edit by KESSLER
%  Changed ChooseRegister&UseRegister to DefAnyregs, added *LMASK and
% *MERGE Cmacros, so the call to useregister will get expanded.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(imports '(opencoded-fns))
 
(define-constant NonTagPart 59)
 
%========================================================================
%   A few useful LISP routines.
%   Select a register.  See if both R1 and R2 are the temporary
%   registers.  If so, then return a register that we can use (i.e.
%   Reg S5, for saving.
 
(de ChooseRegisterFunction (DummyArgument R1 R2)
     (Cond ((And (TempSRegP R1) (TempSRegP R2)) '(Reg 5))
            (T '(Reg T7))))
 
(defanyreg ChooseRegister ChooseRegisterFunction)
%---------------------------------------------------------------------------
%   PickOffIndexedReg will do just that, take in a form
%    (INDEXED (reg xx) 0) and return (reg xx).
 
(de AnyRegPickOffIndexedReg (reg x)    (cadr x))
 
(defAnyReg PickOffIndexedReg AnyRegPickOffIndexedReg)
%---------------------------------------------------------------------------
%   PickOffIndexAmt will extract the y in (INDEXED (reg xx) y).
 
(de AnyRegPickOffIndexAmt (reg x)  (caddr x))
 
(defAnyReg PickOffIndexAmt AnyRegPickOffIndexAmt)
%---------------------------------------------------------------------------
%   Now, we can use either (Reg S5) if both are temp registers, otherwise
%    we use whichever TempRegister is not in use.
 
(de UseRegisterFunction (DummyArgument R1 R2)
     (Cond ((And (TempSRegP R1) (TempSRegP R2))            '(Reg 5))
           ((Or (Equal R1 '(Reg S6)) (Equal R2 '(Reg S6))) '(Reg S7))
           (T      '(Reg S6))))
 
(defAnyReg UseRegister UseRegisterFunction)
%---------------------------------------------------------------------------
%   When called, this will generate the code to save the id and number of
%   arguments into the LinkReg and NargReg for later use.  Note, id's flagged
%   as NoLinkage will not have the instructions generated.
%   This version does an xidloc, instead of the devistating idloc.
 
(compiletime (unfluid '(&linkshift&)))
 
(de MakeLinkRegs(Fn Nargs)
  (cond ((FlagP Fn 'NoLinkage) NIL)
        ((wlessp Nargs 5)
           `((*Move (xIdLoc ,Fn ,Nargs) (Reg 5))
             (*Move (Reg 5) (Reg NargReg))))
        (T `((*Move (xIdLoc ,Fn ,Nargs) (Reg NargReg)))
   )))
 
% These should make xidloc work ok in the assembler.
 
(de Asmxidloc (u)
  (cond ((eq (cadr u) 9) (setq u '(dummy pair 2)))) % ???
  (printexpression
      (plus2 (times2 (eval (list 'idloc (cadr u)))
                     &LinkShift&) (caddr u))))
 
(put 'xidloc 'asmexpressionfunction 'ASMxidloc)
 
% One that should work for lap code.
 
(fluid '(*test-only*))
 
(de binxidloc (u)
  (cond ((not *writingfaslfile)
       (wplus2 (wtimes2 (eval (list 'idloc (cadr u)))
                        &LinkShift&) (caddr u)))
      (t (cond ((not *test-only*)
              (add-next-bit 2)   % Reloc ID, next says to do xidloc.
              (add-next-bit 1))) % (const reloc_id_number))))
         (wplus2 (wtimes2 (findidnumber (cadr u))
                           &LinkShift&) (caddr u)))))
 
(put 'xidloc 'operandvaluefunction 'binxidloc)
