%
% FASLOUT.RED - Top level of fasl file writer
%
% Author:      Eric Benson
%              Symbolic Computation Group
%              Computer Science Dept.
%              University of Utah
% Date:        16 February 1982
% Copyright (c) 1982 University of Utah
%
%  <PSL.COMP>FASLOUT.RED.35, 10-Jun-82 10:41:18, Edit by GRISS
%  Made CompileUncompiledExpressions regular func
%  <PSL.COMP>FASLOUT.RED.12, 30-Apr-82 14:45:59, Edit by BENSON
%  Removed EVAL and IGNORE processing
%  <PSL.COMP>FASLOUT.RED.8, 29-Apr-82 06:23:18, Edit by GRISS
%  moved DEFINEROP call to RLISP-PARSER

%
%  A FASL file will look like this:
%
%       ============================
%       |    Fasl_Magic_Number     |  (currently 99)
%       ============================
%       |  Number of Id Tables     |
%       |==========================|
%       |   Size of Id Table 1     |
%       |   Offset for Table 1     |
%       |--------------------------|
%       |     chars in ID1,1       |
%       |     pname of ID1,1       |
%       |   -------------------    |
%       |           ...            |
%       |   -------------------    |
%       |   # of chars in ID1,j    |
%       |     pname of ID1,j       |
%       |==========================|
%       |           ...            |
%       |==========================|
%       |   Size of Id Table n     |
%       |   Offset for Table n     |
%       |--------------------------|
%       |   # of chars in IDn,1    |
%       |     pname of IDn1        |
%       |   -------------------    |
%       |           ...            |
%       |   -------------------    |
%       |   # of chars in IDn,k    |
%       |     pname of IDn,k       |
%       ============================
%       |  Size of block in items  |
%       |--------------------------|
%       |   Data for fn 1          |
%       |   Code for fn 1          |
%       |   Bit table for fn 1     |
%       |   Init cmds for fn 1     | (may include "execute now")
%       |--------------------------|
%       |          ...             |
%       |--------------------------|
%       |   Data for fn n          |
%       |   Code for fn n          |
%       |   Bit table for fn n     |
%       |   Init cmds for fn n     |
%       ============================
% on syslisp;
% CompileTime
%  internal WConst InstallNow = 1,
%                 ExecuteNow = 2,
%                 DeleteNow  = 3;
% off syslisp;

(compiletime (progn 
 (flag '(codefileheader codefiletrailer allocatefaslspaces) 
       'internalfunction) nil))


(fluid '(*writingfaslfile 
         *lower 
         *needbittable 
         *quiet_faslout))
       

(fluid '(semic* 
         dfprint* 
         uncompiledexpressions* 
         estimatedcompiledsizeofuncexps* 
         modulename* 
         codeout* 
         faslstart*                     % start of memory available ot FASL
         faslblockstart*                % where FASL starts to deposit code
         faslend*                       % end of memory available to FASL
         faslptr*                       % LAP sets its deposit ptr from this
         initfnoffset*                  % offset to first initfn
         bittablebase*                  % where FASL starts to deposit bittbl
         bittableoffset*                % updated by LAP
         maxfasloffset*                 % limit for BitTableOffset!*
         faslfilenameformat* 
         orderedidlists* 
         codeblocksize*                 % bytes of code and literals, etc.
         datablocksize*                 % bytes of psuedo heap data
         faslcodepointerindex*))        % discriminates ***Code**Pointer*** fns
       

(fluid '(maximumfunctionsize*           % 4095 -- make smaller if initfn problems
         idtableoffsets*))              % determined empirically
       

% if initfns are getting too large, this can be made smaller
% (or EstimateCompiledSizeOf can be made more accurate)
(setq maximumfunctionsize* 4095)
(setq estimatedcompiledsizeofuncexps* 0)

(setq idtableoffsets* '((1000 . 1024) (129 . 129)))	% cms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(on syslisp)
(ds hbpsitemsinf (x) (wand 
 x 8388607))
(off syslisp)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de dfprintfasl (u)                     %. Called by TOP-loop, DFPRINT!*
 (prog (nam ty fn *writingfaslfile) 
  (setq *writingfaslfile t) 
  (cond ((atom u) (return nil))) 
  (setq fn (car u)) 
  (cond ((equal fn 'putd) (go db2))) 
  (cond ((not (memq fn '(de df dm dn))) (go db1))) 
  (setq nam (cadr u)) 
  (setq u (cons 'lambda (cddr u))) 
  (setq ty (cdr (assoc fn 
                  '((de . expr) (df . fexpr) (dm . macro) (dn . nexpr))))) 
db3 
  (cond ((equal ty 'macro) 
     (prog (*comp) 
        (putd nam ty u))))              % Macros get defined now
     
  (cond ((flagp nam 'lose) 
  (progn (errorprintf 
               "*** %r has not been defined, because it is flagged LOSE" 
                   nam) 
      (return nil)))) 
  (cond ((flagp ty 'compile) 
   (progn (cond ((and (car uncompiledexpressions*) 	% dump pending exps now
          (neq nam '**fasl**initcode**)) 	% but avoid infinite rec
        (progn (compileuncompiledexpressions) 
            (setq uncompiledexpressions* (cons nil nil))))) 
       (put nam 'cfntype (list ty)) 
       (setq u (cons (list '*entry nam ty (length (cadr u))) (&comproc u nam))) 
       (lap u))) (t 
   %   DepositFaslInfoAfterLap U  >>  % now part of lap
  (saveuncompiledexpression             % should never happen
                (list 'putd (mkquote nam) (mkquote ty) (mkquote u))))) 

  % Use Prin2 since PrintF has problem with linesize
  % Also, suppress printing !*!*FASL!*!*INITCODE!*!*
  %   It clutters output, and CompiledUncompiledExpression will have
  %   already said it's an initfn in a nicer way.
  (cond ((eq nam '**fasl**initcode**) 
       (printf " %p,%p bytes.%n" codeblocksize* datablocksize*)) (t 
  (prin2 (bldmsg " %w %p,%p bytes. " 
                     nam codeblocksize* datablocksize*)))) 
  (return nil) 
db1                                     % Simple S-Expression, maybe Eval it;
  (cond ((not (pairp u)) (return nil))) 
  (cond ((setq fn (get (car u) 'faslpreeval)) 
       (return (apply fn (list u)))) 
  ((and (setq fn (getd (car u))) (equal (car fn) 'macro)) 
       (return (dfprintfasl (apply (cdr fn) (list u)))))) 
  (saveuncompiledexpression u) 
  (return nil) 
db2 
  (setq nam (cadr u)) 
  (setq ty (caddr u)) 
  (setq fn (cadddr u)) 
  (cond ((eqcar nam 'quote) 
  (progn (setq nam (cadr nam)) 
      (cond ((eqcar ty 'quote) 
         (progn (setq ty (cadr ty)) 
             (cond ((and (pairp fn) (member (car fn) '(function quote))) 
             (progn (setq fn (cadr fn)) 
                 (cond ((memq ty '(expr fexpr macro nexpr)) 
                 (progn (setq u fn) 
                     (go db3))))))))))))) 
  (go db1)))
 

(flag '(flag fluid global remflag remprop unfluid) 'eval)

(de faslpreevalloadtime (u) 
 (dfprintfasl (cadr u)))                % remove LOADTIME

% A few things to save space when loading

(de faslpreevalflag (u) 
 (cond ((eqcar (second u) 'quote) 
      (dfprintfasl (cons 'progn (foreach x in (second (second u)) collect 
                            (list 'flag1 (mkquote x) (third u)))))) (t 
 (saveuncompiledexpression u))))

(de faslpreevalfluid (u) 
 (cond ((eqcar (second u) 'quote) 
      (dfprintfasl (cons 'progn (foreach x in (second (second u)) collect 
                            (list 'fluid1 (mkquote x)))))) (t 
 (saveuncompiledexpression u))))

(de faslpreevalglobal (u) 
 (cond ((eqcar (second u) 'quote) 
      (dfprintfasl (cons 'progn (foreach x in (second (second u)) collect 
                            (list 'global1 (mkquote x)))))) (t 
 (saveuncompiledexpression u))))

(de faslpreevaldeflist (u) 
 (cond ((eqcar (second u) 'quote) 
      (dfprintfasl (cons 'progn (foreach x in (second (second u)) collect 
                            (list 'put (mkquote (first x)) (third u) 
                                        (mkquote (second x))))))) (t 
 (saveuncompiledexpression u))))

(de faslpreevalprogn (u) 
 (foreach x in (cdr u) do (dfprintfasl x)))

(de faslpreevallap (u) 
 (cond ((eqcar (cadr u) 'quote) (lap (cadr (cadr u)))) (t 
 (saveuncompiledexpression u))))

(setq uncompiledexpressions* (cons nil nil))

(de saveuncompiledexpression (u) 
 (cond ((not (atom u)) 
  (prog (estimatedcodesize) 
   (setq estimatedcodesize (estimatecompiledsizeof u)) 
   (cond 
          ((lessp (plus estimatedcompiledsizeofuncexps* estimatedcodesize) maximumfunctionsize*) 
   (progn                               %  room for this expression
       (tconc uncompiledexpressions* u) 
       (setq estimatedcompiledsizeofuncexps* 
         (plus estimatedcompiledsizeofuncexps* estimatedcodesize)) 
       (return nil)))) 
   %  no room -- compile existing forms first, then add this one
   (compileuncompiledexpressions) 
   (setq uncompiledexpressions* (cons nil nil)) 
   (tconc uncompiledexpressions* u) 
   (setq estimatedcompiledsizeofuncexps* estimatedcodesize) 
   (return nil)))))
  

% now in 370lap red as DepositFaslInitCommands
% on syslisp;
% lisp procedure DepositFaslInfoAfterLap U;
%  begin scalar Pr, Name, Command;
%   Pr := assoc( '!*Entry, U );
%   Name := cadr Pr;
%   if Name = '!*!*Fasl!*!*InitCode!*!* then
%    <<  @FaslPtr!* := ExecuteNow;
%        FaslPtr!* := FaslPtr!* + 4;
%        @FaslPtr!* := DeleteNow + 16;
%         FaslPtr!* := FaslPtr!* + 4  >>
%   else if not FlagP( Name, 'Internal ) then
%    <<  @FaslPtr!* := InstallNow + 16;
%        FaslPtr!* := FaslPtr!* + 4;
%        @FaslPtr!* := FindIdNumber Name;
%        FaslPtr!* := FaslPtr!* + 4;
%         @FaslPtr!* := FindIdNumber caddr Pr;
%        FaslPtr!* := FaslPtr!* + 4  >>
%  end;
% off syslisp;

(de faslout (fil) 
(progn (setq modulename* fil) 
    (cond ((not *quiet_faslout) 
    (progn (cond ((not (funboundp 'begin1)) 
        (progn (prin2t "FASLOUT: IN files; or type in expressions") 
            (prin2t "When all done execute FASLEND;"))) (t 
        
        (progn (prin2t "FASLOUT: (DSKIN files) or type in expressions") 
            (prin2t "When all done execute (FASLEND)"))))))) 
    (setq codeout* (open (bldmsg faslfilenameformat* modulename*) 
                       'binaryoutput)) 
    (codefileheader) 
    (setq faslcodepointerindex* 0) 
    (setq dfprint* 'dfprintfasl) 
    (setq *writingfaslfile t) 
    (setq *defn t)))

(de faslend nil 
 (cond ((not *writingfaslfile) (stderror "FASLEND not within FASLOUT")) (t 
 
  (progn (compileuncompiledexpressions) 
      (setq uncompiledexpressions* (cons nil nil)) 
      (codefiletrailer) 
      (close codeout*) 
      (setq dfprint* nil) 
      (setq *writingfaslfile nil) 
      (setq *defn nil)))))

(flag '(faslend) 'ignore)

(de comfile (filename) 
 (prog (*defn *writingfaslfile testfile filebase fileext 
              i n dotfound testexts *quiet_faslout) 
  (cond ((idp filename) 
    ((lambda (*lower) (setq filename (bldmsg "%w" filename))) t))) 
  (cond ((not (stringp filename)) (return 
        (nonstringerror filename 'comfile)))) 
  (setq n (isizes filename)) 
  (setq i 0) 
  (while (and (not dotfound) (ileq i n)) 
  (progn (cond ((equal (igets filename i) (char '!.)) (setq dotfound t))) 
        (setq i (iadd1 i)))) 
  (cond (dotfound 
   (progn (cond ((not (filep filename)) 
           (return (conterror 99 "Couldn't find file" (comfile filename)))) (t 
       
       (progn (setq filebase (subseq filename 0 i)) 
           (setq fileext (subseq filename (isub1 i) (iadd1 n)))))))) (t 
  
   (progn (setq testexts '(".build" ".sl" ".red")) 
       (while (and (not (null testexts)) 
             (not (filep (setq testfile (concat filename (first testexts)))))) 
          (setq testexts (cdr testexts))) 
       (cond ((null testexts) 
           (return (conterror 99 "Couldn't find file" (comfile filename)))) (t 
       
       (progn (setq fileext (first testexts)) 
           (setq filebase filename) 
           (setq filename testfile))))))) 
  (errorprintf "*** Compiling %w" filename) 
  (setq *quiet_faslout t) 
  (faslout filebase) 
  (cond ((member fileext '(".build" ".red")) 
       (evin (list filename))) (t 
  (dskin filename))) 
  (faslend) 
  (return t)))
 

(de compileuncompiledexpressions nil 
 (prog (n initfn) 
  (setq n (length (car uncompiledexpressions*))) 
  (cond ((equal n 0) (return nil))) 
  (printf "%f  Init code: %w expression%w" 
          n (cond ((equal n 1) "") (t "s"))) 
  (setq initfn (cons 'progn (car uncompiledexpressions*))) 
  % **Fasl**InitCode** is a special function name for LAP:
  (setq initfnoffset* (difference faslptr* faslblockstart*)) 
  (dfprintfasl (list 'de '**fasl**initcode** 'nil initfn)) 
  (setq uncompiledexpressions* (cons nil nil)) 
  (setq estimatedcompiledsizeofuncexps* 0)))
 

(de estimatecompiledsizeof (exp) 
 (cond ((atom exp) 0) (t 
 % loads
      % call
      % data
      (difference (plus (times addressingunitsperitem (length (cdr exp))) 6 (times addressingunitsperitem (recursiveitemsizeof (cdr exp)))) (times (times 2 addressingunitsperitem) (length (cdr exp)))))))	% overshoot

(de recursiveitemsizeof (exp) 
 (cond ((idp exp) 0) 
 ((inump exp) 0) 
 ((floatp exp) 3) 
 ((numberp exp) 2) 
 ((stringp exp) (plus 3 (quotient (size exp) charactersperword))) 
                                              % errs on the high side
 ((vectorp exp) 
  (prog (sum) 
    (setq sum 0) 
    (for (from i 0 (upbv exp) 1) (do 
    (progn (setq sum (plus sum (recursiveitemsizeof (getv exp i))))))) 
    (return (plus sum 2)))) 
  
 ((atom exp) 2) (t                      % strange case
 (plus 2 (recursiveitemsizeof (car exp)) 	% pairs
        (recursiveitemsizeof (cdr exp))))))

(de codefileheader nil 
(progn (binarywriteitem codeout* (const fasl_magic_number)) 
    (allocatefaslspaces)))

(de allocatefaslspaces nil 
 (prog (b) 
  (setq b (gtwarray nil))               % how much is left?
  (setq b (idifference b (iquotient b 3))) 
  (setq faslend* (gtwarray 0))          % pointer to top of space
  (setq faslstart* (gtwarray b))        % take 2/3 of whatever's left

  (setq orderedidlists* nil)            % Clear orderidlists for each fasl
  (cond (*needbittable                  % split fasl space:
   (progn (setq bittablebase* faslstart*) 	% bit table
       (setq faslblockstart*            % code
        (loc (wgetv faslstart* 
                   (iquotient b 
                              (iquotient (const bittableentriesperword) 
                                         (const addressingunitsperitem)))))))) (t 
       
  (setq faslblockstart* faslstart*)))   % use all fasl space for code

  (setq faslptr* faslblockstart*) 
  (setq maxfasloffset* (idifference faslend* faslptr*)) 
     % MaxFaslOffset!* is used as limit for BitTableOffset!*
  (setq bittableoffset* 0) 
  (foreach offsetpair in idtableoffsets* do 
      (setq orderedidlists* (nconc orderedidlists* 
                                 (list (list (car offsetpair) 	% boundary
                                            (cdr offsetpair) 	% offset
                                            (cdr offsetpair) nil)))))))	% counter
                                            % cdddr = tconc cell
 

% cms dependent
% see version for ASMOUT in 370APRIN.RED
(de findidnumber (u) 
 (prog (i numberhere orderedidlists orderedidlist) 
  (return 
   (cond ((ileq (setq i (idinf u)) 128) i) 
   ((not *writingfaslfile) (id2int u)) 
   ((setq i (get u 'idnumber)) i) (t 
   
    (progn (setq numberhere (id2int u)) 
        (setq orderedidlists orderedidlists*) 
        (while (lessp numberhere (caar orderedidlists)) 
              (setq orderedidlists (cdr orderedidlists))) 
        (setq orderedidlist (car orderedidlists)) 
        (setq i (caddr orderedidlist)) 
        (put u 'idnumber i) 
        (tconc (cdddr orderedidlist) u) 
        (rplaca (cddr orderedidlist) (add1 i)) 
        i))))))
 

(setq macrokludge 'macro)
(putd 'idloc macrokludge '(lambda (x) (findidnumber (cadr x))))

(on syslisp)
(de codefiletrailer nil 
 (prog (s ptr limit) 
  (systemfaslfixup) 

  (binarywriteitem codeout* (length orderedidlists*)) 
  (foreach orderedidlist in orderedidlists* do 
  (progn (binarywriteitem codeout* (sub1 (length (cadddr orderedidlist)))) 
      (binarywriteitem codeout* (cadr orderedidlist)) 	% offset
      (foreach x in (cadddr orderedidlist) do 
      (progn (remprop x 'idnumber) 
          (setq x (strinf (id2string x))) 
          (setq s (wplus2 1 (strlen x))) 
          (setq x (wplus2 x addressingunitsperitem)) 
          (binarywriteblock codeout* x s))) nil)) 
      
  (setq orderedidlists* nil) 
  % the following could be smarter, if 370LAP kept track of space needed
  % this will include space for init fns, which is track of space needed
  (setq s (iquotient (iplus2 (wdifference faslptr* faslblockstart*) 
                          (isub1 addressingunitsperitem)) 
                  addressingunitsperitem)) 
  (binarywriteitem codeout* s)          % how much BPS is needed
  (setq ptr faslblockstart*) 
  (setq limit faslptr*) 
  (while (wlessp ptr limit) 
  (progn (setq s (wtimes2 addressingunitsperitem (hbpsitemsinf (getmem ptr)))) 
      (cond ((or (wlessp s 0) (wgreaterp s 65535)) 
      (progn (delwarray faslstart* faslend*) 
          (stderror (bldmsg 
       "Faslout: Size of code or data block (in addressing units) = %w" 
                           s))))) 
      (binarywriteblock codeout* ptr s) 
      (setq ptr (wplus2 ptr s)))) 
  (binarywriteitem codeout* -1)         % termination code
  (delwarray faslstart* faslend*)))
 
(off syslisp)

(de updatebittable (numberofentries firstentry) 
 (cond (*writingfaslfile 
 (progn (putbittable bittablebase* bittableoffset* firstentry) 
     (setq bittableoffset* (iadd1 bittableoffset*)) 
     (for (from i 2 numberofentries 1) (do 
     (progn (putbittable bittablebase* bittableoffset* 0) 
         (setq bittableoffset* (iadd1 bittableoffset*))))) 
     (cond ((igreaterp bittableoffset* maxfasloffset*) 
        (fatalerror "BPS exhausted during FaslOut; output file too large")))))))
     

%  Defer following to end of file so as not to confuse FASLOUT of this
%  file on systems which already have anonymous fns as FaslPreEval
%  properties.  Otherwise these properties overlay, and then undefined
%  fn calls happen during FASLOUT of this file.
(deflist 
        '((loadtime faslpreevalloadtime) 
        (bothtimes faslpreevalloadtime) 
        (startuptime faslpreevalloadtime) 	% used in kernel
        (deflist faslpreevaldeflist) 
        (flag faslpreevalflag) 
        (fluid faslpreevalfluid) 
        (global faslpreevalglobal) 
        (deflist faslpreevaldeflist) 
        (progn faslpreevalprogn) 
        (lap faslpreevallap)) 
        'faslpreeval)

% of file
