%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           Mips-REGISTERS.SL
% Title:          Mips PSL register allocation                                
% Author:         Winfried Neun , ZIB Berlin 
% Created:        9 Nov 1989
% Status:         Experimental
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%      Define register names
%%
(deflist '((r0 0)(r1 1)(r2 2)(r3 3)(r4 4)(r5 5)(r6 6)(r7 7)
	   (r8 8)(r9 9)(r10 10)(r11 11)(r12 12)(r13 13)(r14 14)
	   (r15 15)(r16 16)(r17 17)(r18 18)(r19 19)(r20 20)
	   (r21 21)(r22 22)(r23 23)(r24 24)(r25 25)(r26 26)
	   (r27 27)(r28 28)(r29 29)(r30 30)(r31 31))
  '*register-code)

(deflist '((f0 0)(f1 1)(f2 2)(f3 3)(f4 4)(f5 5)(f6 6)(f7 7)
	   (f8 8)(f9 9)(f10 10)(f11 11)(f12 12)(f13 13)(f14 14)
	   (f15 15)(f16 16)(f17 17)(f18 18)(f19 19)(f20 20)
	   (f21 21)(f22 22)(f23 23)(f24 24)(f25 25)(f26 26)
	   (f27 27)(f28 28)(f29 29)(f30 30)(f31 31))
  '*f-register-code)

(deflist '((c0 0)(c1 1)(c2 2)(c3 3)(c4 4)(c5 5)(c6 6)(c7 7)
	   (c8 8)(c9 9)(c10 10)(c11 11)(c12 12)(c13 13)(c14 14)
	   (c15 15)(c16 16)(c17 17)(c18 18)(c19 19)(c20 20)
	   (c21 21)(c22 22)(c23 23)(c24 24)(c25 25)(c26 26)
	   (c27 27)(c28 28)(c29 29)(c30 30)(c31 31))
  '*c-register-code)

%%      Other names of registers. (systems view)

(deflist
 '(($0 0) ($1 1) ($2 2) ($3 3) ($4 4) ($5 5) ($6 6) ($7 7) ($8 8)
   ($9 9) ($10 10) ($11 11) ($12 12) ($13 13) ($14 14) ($15 15)
   ($16 16)($17 17)($18 18)($19 19)($20 20)($21 21)($22 22)($23 23)
   ($24 24)($25 25)($26 26) ($27 27)($28 28)($29 29)($30 30)($31 31))
  '*register-code)

%%      Psl friendly names of registers.

(deflist
  '(	(argone r4)(argtwo r5)(argthree r6)(argfour r7)(argfive r8)
	(null r0) (zero r0) (at r1) 
        (retval r2) (t r2) (infmask r18)
        (bndstkptr r20) (bndstklowerbound r21) (bndstkupperbound r22)
        (symfnccenter r23) (symvalcenter r19)
        (ff1 r1) (ff2 r3) (ff3 r15) (ff4 r25)
	(st r29) (sp r29) (unix-st r29) (returnaddr r31)
	(t1 r9) (t2 r10) (t3 r11) (t4 r12) (t5 r13) (t6 r14)
        (t7 r24) ) '*other-name)

(put 'nil '*other-name 'r30) %doesnt work in deflist ??

(deflist
  '(    (heaplast r16) (heaptrapbound r17)
        (bndstkptr r20) (bndstklowerbound r21) (bndstkupperbound r22)
        ) 'Registername)


(de expand-operand (asm-code)
  (cond ((atom asm-code) asm-code)
	(t   (cons (car asm-code)
	       (mapcar (cdr asm-code)
	               (function (lambda(x)(to-standard-name x))))))))

(global '(*REGISTER-NAMES*))

(setq *REGISTER-NAMES*  '((0 r0)(1 r4)(2 r5)(3 r6)(4 r7)(5 r8)))

(de to-standard-name (u)
  (cond ((idp u) (cond ((get u '*other-name) (get u '*other-name))
	                (t u)))
	((pairp u)
	 (cond ((eq (car u) 'REG)
	            (cond ((numberp (cadr u))
	                    (cadr (assoc (cadr u) *REGISTER-NAMES*)))
	                 ((idp (cadr u)) (to-standard-name (cadr u)))
	                 (t (error 0 (list "Unknown Mips register name" u)))))
	       ((member (car u) '(MKITEM WARRAY)) u)
	       ((eq (car u) 'DEFERRED)   (list (to-standard-name (cadr u))))
	       ((member (car u) '(INDEXED DISPLACEMENT))
	        (list (to-standard-name (cadr u))
		      (to-standard-name (caddr u))))
	       ((eq (car u) '+) (cadr u))
	       ((eq (car u) '-) u)
	       (t (mapcar u (function to-standard-name)))))
	(t u)))

(de Mips-register-p (r)
  (cond ((idp r)
	 (cond ((get r 'Registername) T)
               ((get r '*other-name) (Mips-register-p (get r '*other-name)))
	       ((get r '*register-code) t)
	       ((get r '*f-register-code) t)
	       ((get r '*c-register-code) t)
	       (t nil)))
	((eqcar r 'REG) t)
	(t nil)))

(de get-register-code (reg)
      (cond ((pairp reg) (get-register-code (cadr reg)))
	    ((and (get reg '*other-name) 
			(get-register-code (get reg '*other-name))))
            ((eq reg 0) 0)
            ((and (numberp reg) (lessp reg 6)) (plus2 reg 3))
            ((get reg '*register-code) (get reg '*register-code))
	    ((get reg '*f-register-code) (get reg '*f-register-code))
	    ((get reg '*c-register-code) (get reg '*c-register-code))
            (t (error 0 (list "Wrong register argument" reg))))
)
              
(de regp (r) (eqcar r 'reg))

(de fregp (f) (and (eqcar f 'reg) (get (cadr f) '*f-register-code)))
