%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXK:MAIN-START.SL
% Title:          First routine called on startup
% Author:         Brian Beach, Hewlett-Packard CRC
% Created:        16-Feb-84
% Modified:       May 16 1985
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 2-Dec-88 (Winfried Neun)
%  Changed according to new calling sequence etc. (see "ZIB" comments)
% 31-Aug-88 (Julian Padget)
%  Changed fluids to globals, since latter needed in CPSL.
% 28-Oct-87 (Harold Carr)
%  Made panic exit call exit-with-status, rather than redundantely doing
%  what panic exit already did then calling external_exit.
% 13-Apr-87 (Harold Carr & Leigh Stoller)
%  Added external fluid _infbitlength_ to be used in bpsheap.c.
% 26-Sep-86 (Leigh Stoller)
%  Changed exit to unix to be a call to exit library function. The Sun does
%  accept the exit status in (reg 1) like the Bobcat. Called it external_exit
%  so it does not interfere with the cmacro of the same name.
% 22-Sep-86 (Leigh Stoller)
%  Commented out calls 68020_advise which is not defined on the sun.
% 18-Jul-86 (Leigh Stoller)
%  Changed the allocation of argumentblock to be the number of maxregs since
%   it is indexed by base + reg number.
% 15-Feb-85 13:27:53 ()
%  Made it so that PSL allocates the stack it will use.
%  Removed TST.B for stack expansion.
%  Included declaration of STACKSIZE (should go in SYS-CONSTS eventually).
% 9-Jul-84 09:11:55 (RAM)
%  Added fluids (and initialization) unixstdin, unixstdout, unixstderr,
%  unixnull, unixeof, unixtty.
% 3-Jul-84 10:51:27 (RAM)
%  Changed name of initial faslin file to $pnkl/load-psl.b since
%  binaryopenread can now expand shell variable references.
% 29-Jun-84 12:33:38 (RAM)
%  Chagged exit_to_unix to exit-to-unix to match call in kernel-fatal-error.
%  Removed save and restore of A6.
% 31-May-84 08:15:32 (Brian Beach)
%  Changed SYSCLEARIO call to HP_CLEARIO.  SYSCLEARIO is now called by the 
%  loader as part of its startup.
% 16-may-88 (chris burdorf and T. Yamamoto)
%  modified the main entry point so that it could handle our different
%  stack pointer and the sun-4's register windows to get arg c and argv from
%  the operating system.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% we have to overload the macro from $pxc/mips-asm if we want to load
% into 400000 text segment
 
(compiletime (ds kernelstring2string(str)
              (mkitem (tag str) (wshift (inf str) -2  ))))
 
(compiletime (flag '($fluid fluid global $global) 'terminaloperand))
(compiletime    %% Move this hack somewhere else later.
 (defmacro exported-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'exportedsymbol)
         (put id 'symbol id)
         (put id 'vartype 'global)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro internal-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'internalsymbol)
         (put id 'symbol id)
         (put id 'vartype 'global)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro system-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (put id 'vartype 'global)
         ))
   ))

(on fast-integers)

(compiletime (load sys-consts))

(declare-kernel-structure 
 (stack             #.(compiler-constant 'stacksize))
 (tokenbuffer       #.(compiler-constant 'maxtokensize))
 (bndstk            #.(compiler-constant 'bndstksize))
 (catchstack        #.(times2 (compiler-constant 'catchstacksize) 4))
 (hashtable         #.(quotient (compiler-constant 'hash-table-size) 2))
 (onewordbuffer     1)
 (saveargc          1)
 (saveargv          1)
 (datebuffer        1)
 (sparcjumpbuf          6)
 (isasavesystem     2)
 (argumentblock     #.(compiler-constant 'maxargs))
 )

(exported-global
 '(bndstkptr   bndstkupperbound   bndstklowerbound 
   gcarraylowerbound   gcarrayupperbound
   heaplast   oldheaplast
   heaplowerbound  heaptrapbound   oldheaptrapbound   heapupperbound
   _infbitlength_   lastbps   bpslowerbound   mainstartinitialize
   nextbps   oldheapupperbound   oldheaplowerbound
   stacklowerbound   stackupperbound   saveda7
   unixstdin   unixstdout   unixstderr   unixnull   unixeof   unixtty
   ))

(internal-global  '(   symnam  symval  symget symfnc  symprp   ))

(system-global  '(symnam symval symprp  symfnc symget
                  %  stack
                  catchstackptr   catchstack
                  bndstklowerbound   bndstkupperbound
                  bndstkptr   bndstk
                  heap
                  nextsymbol    %% Had SYMBOL as its assembler symbol.  why????
                    % Add these so they are always availbale to compiled code.
                  ***MUST-BE-NIL***   ***MUST-BE-NIL-TOO***
                  *fastcar
                 sparcjumpbuf
   ))

(de init-pointers()

  (setq symnam (wshift symnam -2))
  (setq symval (wshift symval -2))
  (setq symfnc (wshift symfnc -2))
  (setq symprp (wshift symprp -2))
  (setq symget (wshift symget -2))
  (setq hashtable (wshift hashtable -2))
  (setq argumentblock (wshift argumentblock -2))  
  (setq tokenbuffer   (wshift tokenbuffer -2))
  (setq onewordbuffer (wshift onewordbuffer -2))
  (setq datebuffer    (wshift datebuffer  -2))

  (setq catchstackptr (setq catchstack (wshift catchstack -2)))
  (setq bndstklowerbound (setq bndstk (wshift bndstk -2)))
  (setq bndstkupperbound (loc (wgetv bndstk (wdifference bndstksize 1))))
  (setq bndstkptr bndstk)
  (setq heap heaplowerbound)
  (init-viv)
  )

(de init-fluids ()
        % Until problem of labels not appearing is fixed...
  (setq gcarraylowerbound nil)
  (setq gcarrayupperbound nil)
  (setq heaplast nil)
  (setq oldheaplast nil)
  (setq heaptrapbound nil)
  (setq oldheaptrapbound nil)
  (setq heapupperbound nil)
  (setq _infbitlength_ nil)
  (setq lastbps nil)
  (setq bpslowerbound nil)
  (setq mainstartinitialize nil)
  (setq nextbps nil)
  (setq oldheapupperbound nil)
  (setq oldheaplowerbound nil)
  (setq stackupperbound nil)
  (setq unixstdin nil)
  (setq unixstdout nil)
  (setq unixstderr nil)
  (setq unixnull nil)
  (setq unixeof nil)
  (setq unixtty nil)
  (setq ***MUST-BE-NIL*** NIL)
  (setq ***MUST-BE-NIL-TOO*** NIL)
  (setq *fastcar nil))

(compiletime % copying the actual known viv variable to viv area
(progn
 (fluid '(collected-viv*)) 
 
 (dm collect-viv (u) 
   (prog (hhh) 
      (mapobl (function collect-viv-1)) 
      (setq hhh (append `((*entry init-viv expr 0) (*alloc 0)) 
                           (append collected-viv* '((*exit 0)))))
      (return (setq hhh (list 'lap (list 'quote hhh)))))) 
 
 (de collect-viv-1(x) 
    (prog (g) 
      (when (and (or (setq g (get x 'registername))
		     (setq g (get x '*other-name)))
                 (or (get x 'symbol) (get x 'idnumber)))
          (findidnumber x) 
          (return
           (setq collected-viv*
            (append `((*movex ($fluid ,x) (reg ,g))) collected-viv*))))
)) 
))

(collect-viv) %do it now

(compiletime (on mainfound))

(lap '((*entry !p!s!l_!m!a!i!n expr 0)
       (*alloc 1) 

       (*move  (reg argone) (fluid argc))  % system dependent for sun-4. We move
                                       % the argc value (reg i0) into the
                                       %fluid argc. 
       (jal (foreignentry copy_argv))     % call c function to store argv 
                                           % into static space.
       (nop)
       (*wshift (reg retval) -2)           % a word address!
       (*move (reg retval) (fluid argv))
%
% Next step: initializing special registers such that symval and symfnc can be
% accessed by one instruction
%
       (lw (reg t1) ISASAVESYSTEM)
       (la (reg t2) 4711)
       (bne (reg t1) (reg t2) notasavesystem)
       (srl (reg t1) (reg st) 4)
       (*move (Reg t1) ($fluid stacklowerbound)) % is a sys int that way!
       (*move (fluid symval) (reg symvalcenter))
       (*wshift (reg symvalcenter) 2)
       (li (reg t1) 32000)
       (*move (fluid symfnc) (reg symfnccenter))
       (*wshift (reg symfnccenter) 2)
       (*wplus2 (reg symvalcenter) (reg t1))
       (*wplus2 (reg symfnccenter) (reg t1))

       (*move (fluid argc) (reg 1))
       (*move (fluid argv) (reg 2))
       (*move infbitlength (fluid _infbitlength_))
       (*link os_startup_hook expr 2)
        
       (*call init-viv)
       (lw (reg R16) (plus2 isasavesystem 4))
       (*move -1 (reg infmask))
       (*wshift (reg infmask) -5)
       (*move 128 (reg NIL))
       (*mkitem (reg NIL) id-tag)                   % initialize NIL reg
       (*movex ($fluid t) (reg t))
       (*call pre-main)                                 % call PSL
        
       (*move (wconst 0) (reg 1))
       (*linke 1 exit-with-status expr 1)
       (*exit 1)

   notasavesystem
       (sw (reg t2) isasavesystem)
       (srl (reg t1) (reg st) 4)
       (*move (Reg t1) ($fluid stacklowerbound)) % is a sys int that way!
       (*move (fluid symval) (reg symvalcenter))
       (li (reg t1) 32000)
       (*move (fluid symfnc) (reg symfnccenter))
       (*wplus2 (reg symvalcenter) (reg t1))
       (*wplus2 (reg symfnccenter) (reg t1)) 

%*****************************************************************************

       %  Do OS specific initializations (uses argc and argv)
       (*call init-pointers)
       (*move (fluid argc) (reg 1))
       (*move (fluid argv) (reg 2))
       (*move infbitlength (fluid _infbitlength_))
       (*link os_startup_hook expr 2)
       
       (*call init-viv)
       (*move -1 (reg infmask))
       (*wshift (reg infmask) -5)
       (*move 128 (reg NIL))
       (*mkitem (reg NIL) id-tag)                   % initialize NIL reg
       (*movex ($fluid t) (reg t))
       (*call pre-main)                                 % call PSL

panic-exit                      % need to do UNIX cleanup after
                                % a fatal error, so jump here
       (*move (wconst 0) (reg 1))
       (*linke 1 exit-with-status expr 1)
       (*exit 1)

       (*entry exit-with-status expr 1)
         (*alloc 1)
         (*move (reg 1) (frame 1))
         (*link os_cleanup_hook expr 0)
         (*move (frame 1) (reg 1))
         (*link external_exit expr 1)
         (*exit 1)))

(de pre-main ()
  (unixcleario)
  (initialize-symbol-table)
  (initcode)
  (console-print-string (kernelstring2string "About to load LOAD-PSL"))
  (console-newline)
  (faslin (kernelstring2string "load-psl.b"))
  (loader-main))


(compiletime (remflag '($fluid fluid global $global) 'terminaloperand))
(off fast-integers)
