/* 
 * unexec.c - Convert a running program into an COFF file.
 * 
 * Author:	Winfried Neun
 *
 * Synopsis:
 *	unexec( new_name, a_name, data_start, bss_start )
 *	char *new_name, *a_name;
 *	unsigned data_start, bss_start;
 *
 * Takes a snapshot of the program and makes an COFF format file in the
 * file named by the string argument new_name.
 * If a_name is non-NULL, the symbol table will be taken from the given file.
 * 
 * This is a raw version, not many checks are done,etc.
 * Symbols are stripped from file
 *
 */

#include <stdio.h>
#include <filehdr.h>
#include <aouthdr.h>
#include <scnhdr.h>
#include <reloc.h>
#include <linenum.h>
#include <storclass.h>
#include <syms.h>
#include <sys/types.h>
#include <sys/stat.h>

extern char etext, edata;  
#define ETEXT (  ( (unsigned)&etext + (PAGSIZ-1) ) & ~(PAGSIZ-1)  )
#define EDATA (unsigned)&edata

static struct filehdr filhdrold,filhdrnew;
static struct aouthdr aouthdrold,aouthdrnew;
static struct scnhdr scnold,scnnew;

static long STEXT, LNGTEXT;
static long SDATA, LNGDATA;
/* ****************************************************************
 * unexec
 *
 * driving logic
 */

unexec ( nnnnnn_name, o_name, data_start, bss_start )

char *nnnnnn_name, *o_name;
unsigned data_start, bss_start;
{
    FILE *newf;
    int new, a_out = -1;
    char new_name[256], a_name[256];

    /* Expand both filenames now. Place in seperate static arrays so that
       mallocs are not needed at runtime. Messes up the unexec by
       changing the oldbreakvalue.
     */
    strcpy(new_name, expand_file_name(nnnnnn_name));
    if (o_name)
      strcpy(a_name, expand_file_name(o_name));

    if ( a_name && (a_out = open( a_name, 0 )) < 0 )
    {
	perror( a_name );
	return -1;
    }
    if ( (newf = fopen( new_name, "w" )) == NULL )
    {
	perror( new_name );
	return -1;
    }
    fclose(newf);
    new = open( new_name, 1);	    /* This fd will be written with zwrite. */

    if (make_filhdr( new, a_out, data_start, bss_start ) < 0 ||
        make_aouthdr ( new, a_out, data_start, bss_start )  < 0 ||
        make_textscnhdr ( new, a_out, data_start, bss_start )  < 0 ||
        make_datascnhdr ( new, a_out, data_start, bss_start )  < 0 ||
        make_bssscnhdr ( new, a_out, data_start, bss_start )  < 0 ||
	copy_text_and_data( new ) < 0) /* ||
         /*
	copy_sym( new, a_out ) < 0) */
    {
	close( new );
	/* unlink( new_name );	    	/* Failed, unlink new a.out */
	return -1;	
    }

    close( new );
    if ( a_out >= 0 )
	close( a_out );
      mark_x( new_name ); 
    return 0;
}

/* ****************************************************************
 * make_filhdr
 *
 * Make the header in the new a.out from the header in core.
 * Modify the text and data sizes.
 */
static int
make_filhdr( new, a_out, data_start, bss_start )
int new, a_out;
unsigned data_start, bss_start;
{
    unsigned sbrk(), max_addr = sbrk(0);	/* End of allocated space. */

    /* Get symbol table info from header of a.out file if given one. */
    if ( a_out >= 0 )
    {
	if ( read( a_out, &filhdrold, sizeof filhdrold ) != sizeof filhdrold )
	{
	    perror( "Couldn't read header from a.out file" );
	    return -1;
	}

    /*
	if N_BADMAG( hdr )
	{
	    fprintf( stderr, "a.out file doesn't have legal magic number\n" );
	    return -1;
	}
    
	/* Save hdr as it was read in, to find symbol table in a.out file. */
    }
    else
    {
        perror( "symbolfilename* must be specified" );
        return -1;
    }
    /* Construct header from default assumptions.
     * (hdr.a_magic, hdr.a_entry and hdr.a_syms are set above.)
     */
    filhdrnew.f_magic = filhdrold.f_magic;
    filhdrnew.f_nscns = filhdrold.f_nscns;
    filhdrnew.f_timdat =filhdrold.f_timdat;
    filhdrnew.f_symptr = 0;
    filhdrnew.f_nsyms  = 0;
    filhdrnew.f_opthdr =filhdrold.f_opthdr;
    filhdrnew.f_flags = filhdrold.f_flags;
    write (new, &filhdrnew,  sizeof filhdrnew);
}


/* ****************************************************************
 * make_aouthdr
 *
 * Make the header in the new a.out from the header in core.
 * Modify the text and data sizes.
 */
static int
make_aouthdr( new, a_out, data_start, bss_start )
int new, a_out;
unsigned data_start, bss_start;
{
    /* Get symbol table info from header of a.out file if given one. */
    if (read( a_out, &aouthdrold, sizeof aouthdrold ) != sizeof aouthdrold )
        {
            perror( "Couldn't read aout header from coff file" );
            return -1;
        }
    aouthdrnew.magic = aouthdrold.magic;
    aouthdrnew.vstamp = aouthdrold.vstamp;
    aouthdrnew.tsize = aouthdrold.tsize;
    aouthdrnew.dsize = bss_start - aouthdrold.data_start;
    aouthdrnew.bsize = aouthdrold.bsize;
    aouthdrnew.entry = aouthdrold.entry;
    aouthdrnew.text_start = aouthdrold.text_start;
    STEXT = aouthdrold.text_start;
    LNGTEXT = aouthdrold.tsize;
    SDATA   = aouthdrold.data_start;
    LNGDATA = aouthdrnew.dsize;
    aouthdrnew.data_start = aouthdrold.data_start;
    write (new, &aouthdrnew, sizeof aouthdrnew);
}

/* ****************************************************************
 * make_textscnhdr
 *
 * Make the header in the new a.out from the header in core.
 * Modify the text and data sizes.
 */
static int
make_textscnhdr( new, a_out, data_start, bss_start )
int new, a_out;
unsigned data_start, bss_start;
{ 
    /* Get symbol table info from header of a.out file if given one. */ 
    if (read( a_out, &scnold, sizeof scnold ) != sizeof scnold )
        {   
            perror( "Couldn't read text scn header from coff file" ); 
            return -1; 
        }   
    write (new, &scnold , sizeof scnold );
}

 
/* ****************************************************************
 * make_datascnhdr
 *
 * Make the header in the new a.out from the header in core.
 * Modify the text and data sizes.
 */
static int
make_datascnhdr( new, a_out, data_start, bss_start )
int new, a_out;
unsigned data_start, bss_start;
{
    /* Get symbol table info from header of a.out file if given one. */
    if (read( a_out, &scnold, sizeof scnold ) != sizeof scnold )
        {
            perror( "Couldn't read data scn header from coff file" );
            return -1;
        }   
    scnold.s_size = LNGDATA; /*bss_start - data_start ; */
    write (new, &scnold , sizeof scnold ); 
} 

/* ****************************************************************
 * make_bssscnhdr
 * 
 * Make the header in the new a.out from the header in core. 
 * Modify the text and data sizes. 
 */ 
static int 
make_bssscnhdr( new, a_out, data_start, bss_start ) 
int new, a_out; 
unsigned data_start, bss_start; 
{ 
    /* Get symbol table info from header of a.out file if given one. */ 
    if (read( a_out, &scnold, sizeof scnold ) != sizeof scnold ) 
        { 
            perror( "Couldn't read  bss header from coff file" ); 
            return -1; 
        }    
    scnold.s_paddr = bss_start;
    scnold.s_vaddr = bss_start;
    scnold.s_size  = sbrk(0) - bss_start;
    write (new, &scnold , sizeof scnold );  

    if (read( a_out, &scnold, sizeof scnold ) != sizeof scnold )
        {
            perror( "Couldn't read comment header from coff file" );
            return -1;
        }
    write (new, &scnold , sizeof scnold );

}
copy_text_and_data( new )
int new;
    { int result;
      printf ("writing data file length = %x\n",
                            result = write (new, STEXT, LNGTEXT + LNGDATA));
      return (result);}

/* ****************************************************************
 * mark_x
 *
 * After succesfully building the new a.out, mark it executable
 */
static
mark_x( name )
char *name;
{
    struct stat sbuf;
    int um;
 
    um = umask( 777 );
    umask( um );
    if ( stat( name, &sbuf ) == -1 )
    {
        perror ( "Can't stat new a.out" );
        fprintf( stderr, "Setting protection to %o\n", 0777 & ~um );
        sbuf.st_mode = 0777;
    }
    sbuf.st_mode |= 0111 & ~um;
    if ( chmod( name, sbuf.st_mode ) == -1 )
        perror( "Couldn't change mode of new a.out to executable" );

}


