%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Sun Unix 4.2.
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 17-Feb-89 (Chris Burdorf)
%  Use sun3_sigset for os/4.
% 09-Oct-87 (Leigh Stoller)
%  Do not call build-trap-message anymore. It is too flakey to be relied
%   on to tell the truth.
% 31-Apr-87 (Leigh Stoller)
%  After consult with Julian Padget, altered sigunwind to decrease the
%   stack frame by 320 bytes instead of 98. The interrupted address is
%   found 312 bytes into this frame.
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is recieved.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (off lapopt))

(compiletime (flag '(prin2hexa 
                     build-trap-message 
                     sigunwind
                    ) 'internalfunction))

(compiletime
 (progn
   (setf *sigcalls* nil)
   %
   % Create a list of initialization code needed to set up the traps. The
   % return value is the entry points for each signal handler.
   % This is later flaged as a
   %  cmacro so the returned code can be inserted directly into the lap coded
   %  procedure 'initializeinterrupts each time it is called. The list
   %  *sigentries is then inserted to define the entry points.
   (de *sigsetup (signumber function handler errorstring)
       (setf *sigcalls* (append
             % Sigset takes a function pointer.
             `((*move (wconst ,signumber) (reg 1))
               (*move (label ,handler) (reg 2))
               (*link sparc_sigset expr 2))
             *sigcalls*))
       % Return the the function definition for the signal handler.
       `((*entry ,function expr 0)
     ,handler
     (*move (wconst ,signumber) (reg 1))
     (*move (label ,handler) (reg 2))
     (*push (reg a6))
     (*link sigrelse expr 2)
     (*pop (reg a6))
     (*move (fluid nil) (reg nil))
     (*move (quote ,errorstring) (reg 1))
     (*move (reg a6) (reg 2))   %scp pointer
     (*move (reg a0) (reg 3))   % current frame
     (*jcall sigunwind))
       )
 
   % Return the entry point list. Defined as a cmacro.
   (de *sigcall ()
       *sigcalls*)
 
   (defcmacro *sigcall)
   (defcmacro *sigsetup)
 
   )
) % End of compiletime.
 
 
(lap '(
   % (*sigsetup 1  Huphandler  Huphandlerinstruction  "Hup")
       (*sigsetup 2  Inthandler  IntHandlerInstruction  "Interrupt")
       (*sigsetup 3  QuitHandler QuitHandlerInstruction "Quit")
       (*sigsetup 4  IllHandler  IllHandlerInstruction  "Illegal Instruction")
       (*sigsetup 5  Traphandler TrapHandlerInstruction "Trace Trap")
       (*sigsetup 6  IotHandler  IotHandlerInstruction  "IOT Instruction")
       (*sigsetup 7  Emthandler  EmtHandlerInstruction  "EMT Instruction")
       (*sigsetup 8  FpeHandler  FpeHandlerInstruction "Floating Pt Exception")
       (*sigsetup 10 Bushandler  BusHandlerInstruction  "Bus Error")
       (*sigsetup 11 SegHandler  SegHandlerInstruction
                                    "Segmentation Violation")
       (*sigsetup 12 SysHandler  SysHandlerInstruction
                                    "Bad Args to System Call")
       (*sigsetup 13 PipeHandler PipeHandlerinstruction
                                    "Write on Pipe With Noone to Read")
       (*sigsetup 14 AlrmHandler AlrmHandlerInstruction "Alarm Clock")
       (*entry initializeinterrupts-1 expr 0)
       (*sigcall)
       (*exit 0)))
 
(de initializeinterrupts ()
    %  (ieee_handler (strbase (strinf "set"))
    %             (strbase (strinf "common"))
    %             (symfnc (id2int 'fpehandler)))
       (initializeinterrupts-1))
 
(commentoutcode lap
 '((!*entry sigunwind expr 2)
     % At this point, an arg is already set up in register 1 with a message
     % describing the kind of trap.
     % reg 2 contains the ptr to the actual scp.
 
     % (*move (memory (reg st) 312) (reg 2)) % Addr. at which signal occurred
     % (*wplus2 (reg st) 320)                % Pop signal-handling frame
     % (*link build-trap-message expr 2)     % This leaves its result in reg
                                           % 1, so the new message is
     (*jcall stderror)                     % ready to be passed
     ))
 
(de sigunwind(handler scp frm)
     (stderror (build-trap-message 
                    handler 
                    (getmem (wplus2 scp 12)) 
                    (getmem (wplus2 scp 8))
                    frm)))

(fluid '(*dump))

(compiletime
    (ds next-item()
        (progn (prin2hexa (getmem stk)) 
               (console-print-string " ")
               (setq stk (wplus2 4 stk)))))

(de prin2hexa (num)
   (prog (ziff ind)
       (setq ind -28)
     aa
       (setq ziff (wand (wshift num ind) 15))
       (if (wlessp ziff 10) (setq ziff (wplus2 ziff (inf '!0)))
            (setq ziff (wplus2 ziff (wdifference (inf '!a) 10))))
       (unixputc ziff)
       (setq ind (wplus2 4 ind))
       (when (wlessp ind 1) (go aa))
     (return nil)))

(de build-trap-message (trap-type trap-addr stk frm)
    (let (extra-info id (n 0))
       (setq trap-addr (getmem (getmem (wplus2 stk 8))))
       (when *dump 
         (console-newline)
         (console-print-string "************************************")
         (console-newline)

         (console-print-string "*** Signal: ")
         (console-print-string trap-type)
         (console-newline)
         (console-print-string "*** pc: ") (unixputn trap-addr)
         (console-print-string "   sp: ") (unixputn stk)
         (console-newline)(console-newline)

         (console-print-string "*** extended return block at ")
         (unixputn stk) (console-newline)
         (console-print-string "  addr:  ") (next-item)
         (console-print-string "   state: ") (next-item)
         (console-newline)(console-newline)
         (setq stk (wplus2 stk 24))         % 6 words 
         (console-print-string "   A7 - A0:") (console-newline)
         (next-item) (next-item) (next-item) (next-item)
         (console-newline)
         (next-item) (next-item) (next-item) (next-item)
         (console-newline)(console-newline)
         (console-print-string "   S7 - S0:") (console-newline)
         (next-item) (next-item) (console-newline)
         (next-item) (next-item) (console-newline) 
         (next-item) (next-item) (console-newline) 
         (next-item) (next-item) (console-newline) 
         (next-item) (next-item) (console-newline) 
         (next-item) (next-item) (console-newline) 
         (next-item) (next-item) (console-newline) 
         (next-item) (next-item) (console-newline) 

         (console-newline)
         (console-print-string "************************************") 
         (console-newline) )
      (if (funboundp 'code-address-to-symbol)
        (setf extra-info
              " load ADDR2ID for identifying the offending function ")
  
      (progn
           (setq trap-addr (wand trap-addr 16#ffffff))
           (setq id (code-address-to-symbol trap-addr))
           (setf extra-info (bldmsg " in >%w<+0x%x"
                               id
                               (wdifference trap-addr (code-address id)))))
      )
      (bldmsg "%w%w" trap-type extra-info)))
 
