%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% CATCH-THROW - Common Lisp dynamic non-local exits
%
% Author:      Eric Benson
%              Symbolic Computation Group
%              Computer Science Dept.
%              University of Utah
% Date:        12 October 1982
% Copyright (c) 1982 University of Utah
%
% 05-Nov-86 - H. Melenk and W. Neun  , ZIB Berlin
% changed catch and throw to be free of alloc and using vector instrs.
% 05-Feb-86 - W. Neun , ZIB Berlin
%           translated from RLISP to LISP and PSL 3.4 modifications
%           added update-catchstk if the stack or bndstk has been moved
% 10-May-85 - Kessler
% Changed catch and throw so the size of the frame is stored.  This will
% make the Cray be able to restore the top stack frame.
% Corrections for CRAY, MLG, missing ALLOC's
% 03-Mar-83  Nancy Kendzierski
%  Changed declaration of EMSG* from fluid to global.
% Edit by Cris Perdue, 23 Feb 1983 1624-PST
% Modified the stack overflow warning message
% Edit by Cris Perdue, 16 Feb 1983 1032-PST
% Changed catch stack overflow checking to give a continuable error
%  when stack gets low, Reset when all out.
% Edit by Cris Perdue,  4 Feb 1983 1209-PST
% Moved ERRSET to ERROR-ERRORSET from here.
% Edit by Cris Perdue,  3 Feb 1983 1520-PST
% Changed catch stack overflow to talk about the CATCH stack. ()
% Deleted definition of "errset".
%  <PSL.KERNEL>CATCH-THROW.RED.13, 21-Dec-82 15:55:26, Edit by BENSON
%  Added %clear-catch-stack
%  <PSL.KERNEL>CATCH-THROW.RED.13, 16-Dec-82 09:58:59, Edit by BENSON
%  Error not within ErrorSet now causes fatal error, not infinite loop
 
(compiletime (load vector-instructions))
 
(FLUID '(THROWSIGNAL* THROWTAG*))
 
(GLOBAL '(EMSG*))
 
(DM CATCH-ALL (U)
  ((LAMBDA (FN FORMS)
           (LIST (LIST 'LAMBDA '(&&VALUE&&)
 (LIST 'COND
       (LIST 'THROWSIGNAL*
             (LIST 'APPLY FN
              '(LIST THROWTAG* &&VALUE&&)))
       '(T &&VALUE&&)))
                 (CONS 'CATCH (CONS NIL FORMS))))
   (CADR U)
   (CDDR U)))
 
(DM UNWIND-ALL (U)
  ((LAMBDA (FN FORMS)
           (LIST (LIST 'LAMBDA '(&&VALUE&&)
 (LIST 'APPLY FN
       '(LIST (AND THROWSIGNAL* THROWTAG*)
              &&VALUE&&)))
                 (CONS 'CATCH (CONS NIL FORMS))))
   (CADR U)
   (CDDR U)))
 
(DM UNWIND-PROTECT (U)
  ((LAMBDA (PROTECTED_FORM CLEANUP_FORMS)
           (LIST (LIST 'LAMBDA '(&&VALUE&&)
 (CONS (LIST 'LAMBDA '(&&THROWN&& &&TAG&&)
             (CONS 'PROGN CLEANUP_FORMS)
             '(COND (&&THROWN&&
                     (!%THROW &&TAG&&
                              &&VALUE&&))
                    (T &&VALUE&&)))
       '(THROWSIGNAL* THROWTAG*)))
                 (LIST 'CATCH ''$UNWIND-PROTECT$ PROTECTED_FORM)))
   (CADR U)
   (CDDR U)))
 
 
(off R2I)
 
 % This funny definition is due to a PA1FN for CATCH
 
(DF CATCH (U)
  ((LAMBDA (TAG FORMS)
           (CATCH (EVAL TAG) (EVPROGN FORMS)))
   (CAR U)
   (CDR U)))
 
(on R2I) 
 
% Temporary compatibility package.
 
(DM *CATCH (U)
  (CONS 'CATCH (CDR U)))
 
(DE *THROW (X Y)
  (THROW X Y))
 
 % Size is in terms of number of frames
 
(FLUID '(CATCHSTACK CATCHSTACKPTR))
(SETQ CATCHSTACKPTR  (LOC (WGETV CATCHSTACK 0)))
 
(CompileTime
 
(DS  CatchPop() (SETQ CatchStackPtr (LOC (WGETV CatchStackPtr -5))))
)
(compiletime
(ds  CatchStackDecrement (X) (loc (WGETV X -5)))
)
 % Rather large for a smacro, used only from CatchSetupAux /csp
 % Tests structured for fast usual execution /csp
 % Random constant 5 for "reserve" catch stack frames /csp
 
(compiletime
(ds CatchPush (Tag PC SP Env FrmSz)
      (setq  CatchStackPtr (loc (wgetv CatchStackPtr 5)))
      (when (wgeq CatchStackPtr
                 (loc (wgetv CatchStack
                  (wtimes2 (wdifference CatchStackSize 5) 5))))
        (when (weq CatchStackPtr
                 (loc (wgetv CatchStack
                  (wtimes2 (wdifference  CatchStackSize 5) 5))))
          (ContinuableError 99 "Catch-throw stack overflow (warning)" NIL))
        (when (wgeq CatchStackPtr
                 (loc (wgetv CatchStack
                  (wtimes2  CatchStackSize 5))))
          (setf EMSG* "Catch stack overflow")
          (reset)))
     (setf (wgetv CatchStackPtr 0)  Tag)
     (setf (wgetv CatchStackPtr 1)  PC)
     (setf (wgetv CatchStackPtr 2)  SP)
     (setf (wgetv CatchStackPtr 3)  Env)
     (setf (wgetv CatchStackPtr 4)  FrmSz)
))
 
(compiletime (progn
(ds CatchTopTag()   (wgetv CatchStackPtr 0))
 
(ds CatchTagAt (X)  (wgetv  X 0))
 
(ds CatchTopPC()    (wgetv CatchStackPtr 1))
 
(ds CatchTopSP()    (wgetv CatchStackPtr 2))
 
(ds CatchTopEnv()   (wgetv CatchStackPtr 3))
 
(ds CatchTopFrmSz() (wgetv CatchStackPtr 4))
 
% Vector versions
 
(ds VReadCatchStack() (progn (VSetVl 5)
                             (Vload (Vreg 4) Catchstackptr 1)))
 
(ds VCatchTopTag()   (vgetelement (Vreg 4) 0))
 
(ds VCatchTopPC()    (Vgetelement (Vreg 4) 1))
 
(ds VCatchTopSP()    (Vgetelement (Vreg 4) 2))
 
(ds VCatchTopEnv()   (Vgetelement (Vreg 4) 3))
 
(ds VCatchTopFrmSz() (Vgetelement (Vreg 4) 4))
 
(flag '(CatchSetupAux ThrowAux
                      FindCatchMarkAndThrow-error) 'InternalFunction )
 
))
% CatchSetup puts the return address in reg 2, the stack pointer in reg 3
% and calls CatchSetupAux
 
(lap '((*entry CatchSetup expr 1)      %. CatchSetup(Tag)
       (*alloc 0)
       (*MOVE (reg B00) (reg 2))    % Return address of this function
       (*move (reg st) (reg 3))     % Stack into R3
       (*move (reg A3) (reg 4))     % Current Frame Size
       (*wplus2 (reg 3) (reg 4))    % Offset the stack by the future size.
       (*dealloc 0)
       (*JCALL CatchSetupAux)
 )    )
 
(de update-catchstk (diff posi)   % if stack or bndstk have been moved
    (prog (actcatch)      % update sp entries by diff
          (setq actcatch catchstackptr)
    lupe  (setf (wgetv actcatch posi) (wplus2 (wgetv actcatch posi)
                                               diff))
          (setq actcatch (wplus2 actcatch -5))
          (cond ((wgreaterp actcatch catchstack ) (go lupe)))
          (return nil)
)   )
 
(DE CATCHSETUPAUX (TAG PC SP FRMSZ)
  (PROG (PREVIOUS)
        (SETQ PREVIOUS CATCHSTACKPTR)
%%%%%   (CATCHPUSH TAG PC SP (CAPTUREENVIRONMENT) FRMSZ)
        (CATCHPUSH TAG PC SP bndstkptr FRMSZ)
        (SETF THROWSIGNAL* NIL)
        (RETURN PREVIOUS)))
 
(DE !%UNCATCH (PREVIOUS)
  (PROGN (SETQ CATCHSTACKPTR PREVIOUS)
         (SETF THROWSIGNAL* NIL)))
 
(DE !%CLEAR-CATCH-STACK NIL
  (SETQ CATCHSTACKPTR (LOC (WGETV CATCHSTACK 0))))
 
%
% following macros generates code to do the id-to-reg conversion
% for the bindings
% NOTABENE:     only those fluids are bound, that are flagged
%                      *user-variable
 
(compiletime
 (progn
 
(fluid '(collectedregs*))
 
(de collectregs-rewrite (x)
     (when (and (get x 'registername) (flagp x '*user-variable))
      (setq collectedregs*
       (cons `((eq idname (quote ,x)) (setq ,x (bndstkval bndstkptr)))
                                                     collectedregs*))))
 
(dm idtoreg-conversion-rewrite ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-rewrite))
       (cons 'cond (nconc collectedregs*
   `((t (setf (symval (bndstkid bndstkptr))(bndstkval bndstkptr)))) )))
 
)) %end compiletime
 
(DE !%THROW (TAG VALUE)
  (PROG (TopTag Env idname ptr)
        (VReadCatchStack)
        (SETQ TOPTAG (VCatchTOPTAG))
        (when (NOT (OR (NULL TOPTAG)
                       (WEQ TOPTAG '$UNWIND-PROTECT$)
                       (WEQ TAG TOPTAG)))
                              (CATCHPOP)
                              (return (!%THROW TAG VALUE)))
      %% (RESTOREENVIRONMENT (SETQ ENV (VCatchTOPENV)))
         (setq ptr (setq env (VCatchTopenv)))
         (if (wlessp ptr bndstklowerbound)
                        (return (bstackunderflow))
         (while (wgreaterp bndstkptr ptr)
             (setq idname (mkid (bndstkid bndstkptr)))
             (idtoreg-conversion-rewrite)
             (setq bndstkptr (adjustbndstkptr bndstkptr -1))))
 
         (CATCHPOP)
         (SETF THROWSIGNAL* T)
         (SETF THROWTAG* TAG)
         (RETURN (THROWAUX VALUE (VCatchTopPC)
                                 (VCatchTopSP)
                                 (VCatchTopFRMSZ)))))
 
 (lap '((*entry ThrowAux expr 4)
       (*alloc 0) %// Wasted, since ST reset and EXIT resets B00 on CRAY
       (*move (reg 4) (reg a3))
       (*MOVE (reg 3) (reg st))   % restore ST and PC
       (*move (reg 2) (reg a6))
       (*wdifference (Reg st) (Reg a3))    % Take the Stack down
       (*Move (Reg a6) (Reg B00))
       (*Move (Reg st) (Reg a0))           % Set up address
       (blklod (Reg a3) (Reg PrevFrameSize))% Do the load
       (J (reg b00))
 ))
 
(DE THROW (TAG VALUE)
 
 % Throw to $Error$ that doesn't have a catch can't cause a normal error
 % else an infinite loop will result.  Changed to use FatalError instead.
 
 (prog (cstack p)
  (setq cstack catchstack)
  (setq p catchstackptr)
 lupe
  (when (WEQ P CSTACK)
          (if (NOT (WEQ TAG '$ERROR$))
                 (return (FINDCATCHMARKANDTHROW-error tag value))
                 (return (FATALERROR "Error not within ErrorSet"))))
 
  (when (OR (NULL (CATCHTAGAT P)) (WEQ TAG (CATCHTAGAT P)))
                               (return (!%THROW TAG VALUE)))
  (setq p (CATCHSTACKDECREMENT P))
  (go lupe)
))
 
(DE FINDCATCHMARKANDTHROW-error (tag value)
   (CONTERROR 99 "Catch tag %r not found in Throw" TAG
     (THROW TAG VALUE)))
