%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:              PerfMon
% Description:       Interface to Cray XMP performance monitor
% Author:            H. Melenk ,W. Neun       ZIB Berlin
% Created:           17-Mar-87
% Mode:              Lisp
% Package:           Utilities (Cray)
% Status:            Experimental (Do not distribute)
%
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Measuring Performance of PSL 3.4          %
%   using XMP performance monitor           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime
 (progn (de ijksize (x) x) % needed by EX opcode function
        (put 'exchange-jump 'opencode '((Move (Reg S0) (Reg 2))
                                        (EX 000)))
        (load f-vectors)
        (flag '(analyse-subblock) 'internalfunction)
        (setq *syslisp t)))
 
(fluid '(perfmonbuffer* perfmonbufferlength*))
 
(setq perfmonbufferlength* 500)
 
(setq perfmonbuffer* (make-words perfmonbufferlength* 0))
 
(de performance-monitor (type group)
  (prog (x buffer)
     (setq buffer perfmonbuffer*)
     (setq x (cond ((eq type 'on) 0)
                   ((eq type 'off) 2)
                   ((eq type 'report) 1)
                   ((eq type 'reset) 3)
                   (t 17))) % let it happen!
     (wputv buffer 0 x)
     (wputv buffer 1 group)
     (wputv buffer 2 perfmonbufferlength*)
     (exchange-jump (inf perfmonbuffer*) 8#107)
)  )
 
(de analyse-performance ()
  (prog (a inc end buffer)
   (setq buffer (inf perfmonbuffer*))
   (performance-monitor 'report 0 buffer
                                  perfmonbufferlength*)
   (setq a (wgetv buffer 4))
   (setq end (wplus2 (wgetv buffer 3) -1))
   (setq inc (wgetv buffer 11))
   (for (from i a end inc)
      (do (analyse-subblock (wplus2 buffer i))))))
 
(de analyse-subblock (adr)
     (terpri)
     (prin2t (list "** User Task number:" (wgetv adr 0)))
     (prin2t (list "** Latest Group was:" (wgetv adr 1)))
     (terpri)
        (prin2T "Number of:")
        (prin2t (list "Instruction issued    " (wgetv adr 2)))
        (prin2t (list "CP's holding issue    " (wgetv adr 3)))
        (prin2t (list "Fetches               " (wgetv adr 4)))
        (prin2t (list "I/O references        " (wgetv adr 5)))
        (prin2t (list "CPU references        " (wgetv adr 6)))
        (prin2t (list "Float.point add ops   " (wgetv adr 7)))
        (prin2t (list "Float.point multiplies" (wgetv adr 8)))
        (prin2t (list "Float.point recips    " (wgetv adr 9)))
        (prin2t (list "** Total group 0 **   " (wgetv adr 10)))
     (terpri)
        (prin2T "Hold Issue Conditions:")
        (prin2t (list "Semaphores            " (wgetv adr 11)))
        (prin2t (list "Shared registers      " (wgetv adr 12)))
        (prin2t (list "A registers           " (wgetv adr 13)))
        (prin2t (list "S registers           " (wgetv adr 14)))
        (prin2t (list "V registers           " (wgetv adr 15)))
        (prin2t (list "V functional units    " (wgetv adr 16)))
        (prin2t (list "Scalar memory         " (wgetv adr 17)))
        (prin2t (list "Block memory          " (wgetv adr 18)))
        (prin2t (list "** Total group 1 **   " (wgetv adr 19)))
     (terpri)
        (prin2T "Number of:")
        (prin2t (list "Fetches               " (wgetv adr 20)))
        (prin2t (list "Scalar references     " (wgetv adr 21)))
        (prin2t (list "Scalar conflicts      " (wgetv adr 22)))
        (prin2t (list "I/O references        " (wgetv adr 23)))
        (prin2t (list "I/O conflicts         " (wgetv adr 24)))
        (prin2t (list "Block references      " (wgetv adr 25)))
        (prin2t (list "Block conflicts       " (wgetv adr 26)))
        (prin2t (list "Vector memory refs.   " (wgetv adr 27)))
        (prin2t (list "** Total group 2 **   " (wgetv adr 28)))
     (terpri)
        (prin2T "Number of Instructions:")
        (prin2t (list "Channel & Jump        " (wgetv adr 29)))
        (prin2t (list "Scalar                " (wgetv adr 30)))
        (prin2t (list "Vector logic & integer" (wgetv adr 31)))
        (prin2t (list "Vector float          " (wgetv adr 32)))
        (prin2t (list "Vector memory         " (wgetv adr 33)))
        (prin2t (list "Vector integer        " (wgetv adr 34)))
        (prin2t (list "Vector float          " (wgetv adr 35)))
        (prin2t (list "Vector memory refs.   " (wgetv adr 36)))
        (prin2t (list "** Total group 3 **   " (wgetv adr 37))))
 
